/**
 * 
 */
package ca.mcgill.cs.swevo.diffcat.view;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.Viewer;

import ca.mcgill.cs.swevo.semdiff.core.DetectorResults;
import ca.mcgill.cs.swevo.semdiff.core.Result;
import ca.mcgill.cs.swevo.semdiff.core.Transaction;
import ca.mcgill.cs.swevo.semdiff.ui.dialogs.FilterLabelProvider;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.ListUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.model.DiffCodes;
import ca.mcgill.cs.swevo.diffcat.util.CastConverter;
import ca.mcgill.cs.swevo.diffcat.view.NodeSorter.DiffCatComparator;
import ca.mcgill.cs.swevo.diffcat.view.filters.DiffFilterChainBuilder;

/**
 * Content provider for the CategoryView
 * 
 * @author david
 *
 */
public class CategoryViewContentProvider implements ITreeContentProvider {
	
	private final CategoryView view;
	private final DiffFilterChainBuilder diffFilterBuilder;
	private final CastConverter<Result, DiffCatResult> conv;
	private final DiffCatComparator sorter;
	
	public CategoryViewContentProvider (CategoryView view) {
		this.view = view;
		diffFilterBuilder = new DiffFilterChainBuilder();
		conv = new CastConverter<Result, DiffCatResult>();
		sorter = new DiffCatComparator();
	}
	
	public Object[] getChildren(Object parentElement) {
		if (parentElement instanceof TreeParent) {
			return ((TreeParent) parentElement).getChildren();
		}
		return new Object[0];
	}

	public Object getParent(Object element) {
		if (element instanceof TreeObject) {
			return ((TreeObject) element).getParent();
		}
		return null;
	}

	public boolean hasChildren(Object element) {
		if (element instanceof TreeParent) {
			TreeParent p = (TreeParent) element;
			if (p.getChildren().length > 0) {
				return true;
			}
		}
		return false;
	}

	public Object[] getElements(Object inputElement) {
		
		if (inputElement != null) {
			Transaction input = (Transaction) inputElement;
			DetectorResults detectorResults = input.getResultsMap().get("ca.mcgill.cs.swevo.diffcat");
			if (detectorResults == null) {
				return new Object[0];
			}
			else if (detectorResults.getResults() == null) {
				return new Object[0];
			} 
			else  {
				Collection<DiffCatResult> diffCatResults = getVisibleDiffs(input);
				Set<ClassSigNode> hierarchy = buildHierarchy(diffCatResults);
				return hierarchy.toArray();
			}
		} else {
			return null;
		}
	}

	public void dispose() {
	}

	public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		
	}
	
	private Set<IFilter<DiffCatResult>> getDiffCatFilters () {
		Set<IFilter<DiffCatResult>> filters = new HashSet<IFilter<DiffCatResult>>();
		filters.add(diffFilterBuilder.makePreferenceFilter());
		filters.addAll(view.getAdHocFilters());
		return filters;
	}
	
	private Set<ClassSigNode> buildHierarchy (Collection<DiffCatResult> results)
	{
		Map<ClassSigNode, Map<String, TreeParent>> subMemberMap = new HashMap<ClassSigNode, Map<String,TreeParent>>();
		
		Set<ClassSigNode> parents = new HashSet<ClassSigNode>();
		for (DiffCatResult result : results) {
			
			String changeType = result.getChangeType();
			if (changeType.equals(ChangeOperation.CLASS_DELETE.name())) {
				ClassSigNode node = new ClassSigNode(result.getLeftClassSignature(), result.getPathFrom());
				parents.add(node);
				node.addChild(new DiffCatNode(result));
				continue;
			}
			else if (changeType.equals(ChangeOperation.CLASS_INSERT.name())) {
				ClassSigNode node = new ClassSigNode (result.getRightClassSignature(), result.getPathTo());
				parents.add(node);
				node.addChild(new DiffCatNode(result));
				continue;
			}
			
			String parentSig = result.getRightClassSignature();
			String rightPath = result.getPathTo();
			String methodSig = result.getRightMethodSignature();
			String fieldSig = result.getRightFieldSignature();
			ClassSigNode chosenParent = null;
			
			for (ClassSigNode parent : parents) {
				if (parentSig == null) {
					if (parent.getName().equals("MISC")) {
						chosenParent = parent;
					}
				}
				else {
					if (parent.getName().equals(parentSig) && parent.getPath().equals(rightPath)) {
						chosenParent = parent;
					}
				}
			}
			
			if (chosenParent == null) {
				if (parentSig == null) {
					chosenParent = new ClassSigNode ("MISC", "");
				}
				else {
					chosenParent = new ClassSigNode(parentSig!=null?parentSig:"MISC", rightPath);
				}
				parents.add(chosenParent);
				subMemberMap.put(chosenParent, new HashMap<String, TreeParent>());
			}
			
			// high-level diff
			if (isHighLevelDiff(result)) {
				chosenParent.addChild(new DiffCatNode(result));
				continue;
			}
			else if (fieldSig == null && methodSig == null) {
				// class-level document update
				if (DiffCodes.DOCUMENT_DIFF.passesFilter(result)) {
					chosenParent.addChild(new DiffCatNode(result));
					continue;
				}
			}
			
			Map<String, TreeParent> memberMap = subMemberMap.get(chosenParent);
			TreeParent node = null;
			if (memberMap.containsKey(methodSig)) {
				node = memberMap.get(methodSig);
			}
			else if (memberMap.containsKey(fieldSig)) {
				node = memberMap.get(fieldSig);
			}
			else if (methodSig != null) {
				node = new MethodSigNode(methodSig);
				memberMap.put(methodSig, node);
				chosenParent.addChild(node);
			}
			else if (fieldSig != null) {
				node = new FieldSigNode(fieldSig);
				memberMap.put(fieldSig, node);
				chosenParent.addChild(node);
			}
			else {
				continue;
			}
			
			node.addChild(new DiffCatNode(result));
		}
		
		return parents;
	}
	
	private boolean isHighLevelDiff (DiffCatResult r) {
		String changeType = r.getChangeType();
		if (changeType.equals(ChangeOperation.FUNCTIONALITY_DELETE.name()) || 
			changeType.equals(ChangeOperation.FUNCTIONALITY_INSERT.name()) ||
			changeType.equals(ChangeOperation.TRIVIAL_CONSTRUCTOR_DELETE.name()) ||
			changeType.equals(ChangeOperation.TRIVIAL_CONSTRUCTOR_INSERT.name()) ||
			changeType.equals(ChangeOperation.ATTRIBUTE_DELETE.name()) || 
			changeType.equals(ChangeOperation.ATTRIBUTE_INSERT.name()) || 
			changeType.equals(ChangeOperation.UNCLASSIFIED.name()) || 
			DiffCodes.CLASS_SIGNATURE_DIFF.passesFilter(r)) 
		{
			return true;
		}
		
		return false;
	}
	
	public Collection<DiffCatResult> getVisibleDiffs (Transaction input) {
		DetectorResults detectorResults = input.getResultsMap().get("ca.mcgill.cs.swevo.diffcat");
		if (detectorResults == null) {
			return Collections.emptyList();
		}
		else if (detectorResults.getResults() == null) {
			return Collections.emptyList();
		} 
		else  {
			List<DiffCatResult> diffCatResults = conv.downCast(detectorResults.getResults());
			Set<IFilter<DiffCatResult>> filters = getDiffCatFilters();
			return CollectionUtil.filterAndCopy(diffCatResults, filters);
		}
	}
	
	public String getLeftPath (TreeParent parent) {
		DiffCatResult child = getAnyChild(parent);
		if (child == null) {
			return null;
		}
		return child.getPathFrom();
	}
	
	public String getRightPath (TreeParent parent) {
		DiffCatResult child = getAnyChild (parent);
		if (child == null) {
			return null;
		}
		return child.getPathTo();
	}
	
	public LRTuple<LRTuple<Integer>> getPositionsOf (TreeParent parent) {
		List<DiffCatResult> children = getAllChilds(parent);
		if (children.isEmpty()) {
			return new LRTuple<LRTuple<Integer>>(null, null);
		}
		Collections.sort(children, sorter);
		DiffCatResult result = (DiffCatResult) children.get(0);
		
		int posFrom = result.getPosFrom();
		int posTo = result.getPosTo();
		
		int lenFrom = result.getLenFrom();
		int lenTo = result.getLenTo();
		
		return new LRTuple<LRTuple<Integer>>(new LRTuple<Integer>(posFrom, lenFrom), new LRTuple<Integer>(posTo, lenTo));
	}
	
	private List<DiffCatResult> getAllChilds (TreeParent parent) {
		List<DiffCatResult> list = new ArrayList<DiffCatResult>();
		TreeObject[] children = parent.getChildren();
		if (children == null) {
			return list;
		}
		for (TreeObject o : children) {
			if (o instanceof DiffCatNode) {
				list.add(((DiffCatNode) o).getResult());
			}
			else if (o instanceof TreeParent) {
				list.addAll(getAllChilds((TreeParent) o));
			}
		}
		
		return list;
	}
	
	private DiffCatResult getAnyChild (TreeParent parent) {
		TreeObject[] children = parent.getChildren();
		if (children == null) { 
			return null; 
		}
		for (int i = 0; i < children.length; i++) {
			TreeObject child = children[i];
			if (child instanceof DiffCatNode) {
				return ((DiffCatNode) child).getResult();
			}
			else if (child instanceof TreeParent) {
				DiffCatResult r = getAnyChild((TreeParent) child);
				if (r != null) {
					return r;
				}
			}
		}
		
		return null;
	}
}