package ca.mcgill.cs.swevo.diffcat.view;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.eclipse.compare.CompareEditorInput;
import org.eclipse.compare.ICompareNavigator;
import org.eclipse.compare.INavigatable;
import org.eclipse.compare.IPropertyChangeNotifier;
import org.eclipse.compare.contentmergeviewer.TextMergeViewer.DiffCatHackNavi;
import org.eclipse.compare.internal.CompareEditorInputNavigator;
import org.eclipse.compare.internal.merge.DocumentMerger.Diff;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;


import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult; 
import ca.mcgill.cs.swevo.diffcat.view.diff.DiffShell;
import ca.mcgill.cs.swevo.diffcat.view.filters.BestGuessShellFinder;
import ca.mcgill.cs.swevo.diffcat.view.filters.BestRightGuessShellFinder;
import ca.mcgill.cs.swevo.diffcat.view.filters.EnclosesShellFilter;
import ca.mcgill.cs.swevo.diffcat.view.filters.InRangeOfShellFinder;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.LRTupleListMap;
import ca.mcgill.cs.swevo.util.LRTupleMap;
import ca.mcgill.cs.swevo.util.ListUtil;

/**
 * Used to programmatically navigate diffs in a TreeViewer. 
 * 
 * @author david
 *
 */
public class TreeViewNavigator implements IPropertyChangeListener {

	private LRTupleListMap<String, DiffShell> editorDiffs;
	private LRTupleMap<String, ICompareNavigator> navigators;
	
	public TreeViewNavigator () {
		editorDiffs = new LRTupleListMap<String, DiffShell>();
		navigators = new LRTupleMap<String, ICompareNavigator>();
	}
	
	/**
	 * Navigates the given editor input to the most appropriate diff
	 * for the result.
	 */
	public boolean navigateToDiff (CompareEditorInput input, DiffCatResult result)
	{ 
		String from = result.getPathFrom();
		String to   = result.getPathTo(); 
		
		// always cache diffs if possible, so can get position info once editor is open
		if (! editorDiffs.containsKey(from, to)) {
			List<DiffShell> shells = initDiffs(input);
			editorDiffs.addAll(from, to, shells);
			navigators.put(from, to, input.getNavigator());
			input.addPropertyChangeListener(this);
		}
		
		// our memoization of the diff positions
		List<DiffShell> diffs = editorDiffs.get(from, to);
		
		if (diffs.size() == 0) {
			return true;
		}
		
		// get the appropriate diff for the result
		InRangeOfShellFinder inRangeFinder = new InRangeOfShellFinder(result);
		int rangeIndex = ListUtil.firstIndexOf(diffs, inRangeFinder);

		//navigate to the diff using range
		if (rangeIndex != -1) {
			navigateToDiff( rangeIndex, navigators.get(from, to));
			return true;
		}
		
		EnclosesShellFilter enclosesFinder = new EnclosesShellFilter(result);
		int enclIndex  = ListUtil.firstIndexOf(diffs, enclosesFinder);
		
		// navigate to the diff using enclose
		if (enclIndex != -1) {
			navigateToDiff (enclIndex, navigators.get(from, to));
			return true;
		}
		
		// else, find the closest diff (using both sides and only the right side)
		BestGuessShellFinder bestGuessFinder = new BestGuessShellFinder(result);
		CollectionUtil.applyToAll(diffs, bestGuessFinder);
		DiffShell bestGuess = bestGuessFinder.getBestGuess();
		int bestGuessIndex = diffs.indexOf(bestGuess);
		
		if (bestGuessIndex != -1) {
			navigateToDiff (bestGuessIndex, navigators.get(from, to));
		}
		
		// 
		return false;
	}
	
	private void navigateToDiff (int index, ICompareNavigator navigator) {
		navigateToStart(navigator);
		for (int i = 0; i < index; i++) {
			navigator.selectChange(true);
		}
	}
	
	private void navigateToStart (ICompareNavigator navigator) {
		while (! navigator.selectChange(false));
	}
	
	private List<DiffShell> initDiffs (CompareEditorInput input)
	{
		CompareEditorInputNavigator navigator = input.getNavigator();
		navigateToStart(navigator);
		
		DiffCatHackNavi diffCat = getDiffCatNavigator(input);
		
		ArrayList<DiffShell> shells = new ArrayList<DiffShell>();
		
		boolean stop = false;
		while (! stop) {
			
			DiffShell current = mapToShell(diffCat.getFather().getfCurrentDiff());
			
			shells.add(current);
			
			stop = navigator.selectChange(true);
		}
		
		//
		return shells;
	}
	
	private DiffShell mapToShell (Diff diff) {
		Position l = diff.getLeftPos();
		Position r = diff.getRightPos();
		return new DiffShell(l.offset, l.length, r.offset, r.length);
	}
	
	private DiffCatHackNavi getDiffCatNavigator (CompareEditorInput input)
	{
		CompareEditorInputNavigator navigator = input.getNavigator();
		INavigatable[] navigatables = navigator.getNavigatables();
		for (int i = 0; i < navigatables.length; i++) {
			INavigatable nav = navigatables[i];
			if (nav instanceof DiffCatHackNavi) {
				return (DiffCatHackNavi) nav;
			}
		}
		return null;
	}

	@Override
	public synchronized void propertyChange(PropertyChangeEvent event) {
		if (event.getProperty().equals(Activator.DELETED_EVENT)) {
			Object source = event.getSource();
			CompareEditorInput input = (CompareEditorInput) source;	// XXX we know this
			Set<LRTuple<String>> keySet = new HashSet<LRTuple<String>>(navigators.keySet());
			for (LRTuple<String> key : keySet) {
				ICompareNavigator iCompareNavigator = navigators.get(key.getLeft(), key.getRight());
				if (iCompareNavigator == null) {
					continue;
				}
				if (input.getNavigator() == iCompareNavigator) {
					navigators.remove(key.getLeft(), key.getRight());
					editorDiffs.remove(key.getLeft(), key.getRight());
				}
			}
		}
	}
}
