/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.eclipse.jdt.core.ITypeRoot;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.PPAASTParser;
import org.eclipse.jdt.core.dom.SimpleName;

import ca.mcgill.cs.swevo.ppa.PPAASTUtil;
import ca.mcgill.cs.swevo.ppa.PPAOptions;
import ca.mcgill.cs.swevo.ppa.ui.PPAUtil;
import ca.mcgill.cs.swevo.semdiff.core.Detector;
import ca.mcgill.cs.swevo.semdiff.core.DetectorInput;
import ca.mcgill.cs.swevo.semdiff.core.DetectorResults;
import ca.mcgill.cs.swevo.semdiff.core.Repository;
import ca.mcgill.cs.swevo.semdiff.core.Result;
import ca.mcgill.cs.swevo.semdiff.core.SemDiffJavaSource;
import ca.mcgill.cs.swevo.semdiff.core.Transaction;
import ca.mcgill.cs.swevo.semdiff.core.util.PPAASTParserWorker;
import ca.mcgill.cs.swevo.util.CollectionMap;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.ListUtil;
import ca.mcgill.cs.swevo.util.SetUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;
import ca.mcgill.cs.swevo.util.io.IOUtils;

import ca.mcgill.cs.swevo.diffcat.analysis.ClassRenameDetector;
import ca.mcgill.cs.swevo.diffcat.cd.ChangeDistillerProxy;
import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.reconciler.ClassInsDelReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.ConstructorInsDelReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.FieldRenameReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.LocalVarExtractionReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.ParamRenameReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.ParentClassInsDelReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.ParentInterfaceInsDelReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.RenameReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.SmallMethodReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.ThrownExceptionsReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.TypeRenameReconciler;
import ca.mcgill.cs.swevo.diffcat.reconciler.VarExtractionReconciler;
import ca.mcgill.cs.swevo.diffcat.refiner.CastExpressionRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.IDiffRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.IfConditionUpdateRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.InitRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.ReturnStatementRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.StringCharInterchangeRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.StringSplitArgRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.SuperConstructorRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.ThisExpressionRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.TypeUpdateRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.VariableDeclarationRefiner;
import ca.mcgill.cs.swevo.diffcat.util.RenameHelper;
import ca.mcgill.cs.swevo.diffcat.util.SemDiffFilePathComparer;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.diffcat.util.filter.BasicEqualityFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsAffectedByRenamesFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsClassDeleteFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsClassInsertDeleteFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsClassInsertFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsClassRenameFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsRenameFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsStatementBasedFilter;
import ca.mcgill.cs.swevo.diffcat.util.op.ResetNodesOp;
import ca.mcgill.cs.swevo.diffcat.util.op.SetPathsOp;
import ca.mcgill.cs.swevo.diffcat.util.op.SignatureReconcileOp;
import ca.mcgill.cs.swevo.diffcat.util.op.TransactionIdOp;

/**
 * Main start point of DiffCat. Coordinates the entire analysis. 
 * @author dkawry
 */
public class MainController implements Detector {
 
	/**
	 * Utility for finding changed file pairs
	 */
	private SemDiffFilePathComparer comparer;
	
	/**
	 * Reused diffing components
	 */
	private ChangeDistillerProxy changeDistiller;
	
	/**
	 * For Rename Ops
	 */
	private RenameHelper renameHelper;
	
	/**
	 * For class rename detection
	 */
	private ClassRenameDetector classRenameDetector;
		
	public MainController () {
		comparer 				= new SemDiffFilePathComparer();
		changeDistiller 		= ChangeDistillerProxy.getProxy();
		renameHelper			= new RenameHelper();
		classRenameDetector 	= new ClassRenameDetector();
	}
		
	/**
	 * Takes as input a bunch of old file versions and a bunch of new file versions and uses Diffcat's main algo to 
	 * compute the diffs between the file versions. We require maps that map file paths to the compilation units at 
	 * those file paths. We need the file paths to distinguish different files that are identical at the file level 
	 * (this happens a lot in Ant). 
	 * 
	 * DiffCat first associates together left/right files with identical file paths. Others are then associated via 
	 * rename detection. 
	 *  
	 * @param leftFiles maps file paths to their compilation units (old files)
	 * @param rightFiles maps file paths to their compilation units (new files)
	 * @return the diffcat results for those files. 
	 */
	public Set<DiffCatResult> getStructuralDiffs (Map<String, CompilationUnit> leftFiles, Map<String, CompilationUnit> rightFiles)
	{
		Map<String, SemDiffJavaSource> version1ast = convertMap(File.separator, leftFiles);
		Map<String, SemDiffJavaSource> version2ast = convertMap(File.separator, rightFiles);
		
		return doGetStructuralDiffs(version1ast, version2ast);
	}

	public Set<DiffCatResult> doGetStructuralDiffs(
			Map<String, SemDiffJavaSource> version1ast,
			Map<String, SemDiffJavaSource> version2ast) 
	{
		Map<String, List<Result>> resultMap = new HashMap<String, List<Result>>();
		DetectorInput input = new DetectorInput(version1ast, version2ast, resultMap);
		
		Transaction trxn = new Transaction();
		trxn.setId(-1);
		
		Repository repo = new Repository();
		
		DetectorResults results = detect(repo, trxn, input);
		List<Result> list = results.getResults();
		
		Set<DiffCatResult> set = new HashSet<DiffCatResult>();
		for (Result r : list) {
			set.add((DiffCatResult) r);
		}
		
		return set;
	}
	
	private Map<String, SemDiffJavaSource> convertMap (String fs, Map<String, CompilationUnit> files) {
		Map<String, SemDiffJavaSource> map = new HashMap<String, SemDiffJavaSource>();
		for (String path : files.keySet()) {
			map.put(fs + path, new SemDiffJavaSource(null, files.get(path)));
		}
		return map;
	}
	
	@Override
	public DetectorResults detect(Repository repository, Transaction transaction, DetectorInput input) {
		
		Set<DiffCatResult> renames ;
		Set<DiffCatResult> changes ;
		Set<DiffCatResult> insDelClasses ;
		Set<DiffCatResult> phase1Centric ;
		
		// Step 1. PPA is already done for us by SemDiff
		
		// Step 2. Get changes from ChangeDistiller for modified files 
		changes = getInitialChangesForModifiedFiles(input);
		
		/* 
		 * Find class ins-del pairs that are actually renames.
		 * Remove class ins-del rename pairs and replace with rename
		 * results. Otherwise keep ins-del results  
		 */
		insDelClasses = getAddedAndDeletedClasses(input);
		findClassRenames(changes, insDelClasses);
		Set<DiffCatResult> classRenames = SetUtil.filterAndCopy(insDelClasses, new IsClassRenameFilter());
		
		//  
		renames = new HashSet<DiffCatResult> ();
		
		// Store the class renames in a separate pool. 
		// We will no longer be able to find them after rename rollback
		// and need to keep track of them to return them to users
		renames.addAll (classRenames);
		
		// Remove rename results from insert delete results
		insDelClasses = SetUtil.filterAndCopy(insDelClasses, new IsClassInsertDeleteFilter());
		
		// Find diffs for renamed classes and add to central diff pool
		changes.addAll (getChangesForRenamedFiles(classRenames));
		
		// Step 3. Refine diffs to find local var, field, and parameter renames
		addDiffCatDetectedRenames(changes); 
		
		// Save all renames for later. (rename rollback eliminates them) 
		renames.addAll (SetUtil.filterAndCopy(changes, new IsRenameFilter()));
		
		// Find all changes that might disappear after rename rollback
		changes = findRenameAffectedChanges(changes);
		phase1Centric = SetUtil.filterAndCopy(changes, new IsAffectedByRenamesFilter());
		phase1Centric.addAll(SetUtil.filterAndCopy(changes, new IsStatementBasedFilter()));
		
		// Undo renames 
		undoRenames(renames, input);
		
		// Get diffs once again
		changes = getInitialChangesForModifiedFiles(input);			// modified files
		changes.addAll (getChangesForRenamedFiles (classRenames));	// renamed files
		changes.addAll (insDelClasses);								// added and deleted classes
		changes.addAll (renames);									// renames no longer reported
		
		// We might find additional renames now 
		addDiffCatDetectedRenames(changes);
		
		// find local variable extractions
		changes = doLocalVarExtraction(changes);
		
		// Add results that were no longer found
		for (DiffCatResult change : phase1Centric) {
			boolean processedInPhase2 = processedInPhase2 (change, changes);
			if (! processedInPhase2) {
				// not entirely accurate. could have also disappeared
				// because of variable extractions. but w/e
				change.setChangeType(ChangeOperation.AFFECTED_ONLY_BY_RENAMES.name());	
				changes.add(change);
			}
			
		}
		
		// Refine and reconcile the diffs
		changes = refineAndReconcileDiffs(changes);
		
		// Redo renames
		redoRenames(renames, input);
		
		// set transaction Ids and other dater
		int id = transaction.getId();
		changes = SetUtil.apply(changes, new TransactionIdOp(id));
		changes = SetUtil.apply(changes, new ResetNodesOp());			// to get the nodes to print properly
		changes = SetUtil.apply(changes, new SignatureReconcileOp());	// to fix some signature paths
		
		return new DetectorResults(new LinkedList<Result>(changes)); 
	}

	private Set<DiffCatResult> getInitialChangesForModifiedFiles (DetectorInput input)
	{
		Map<String, SemDiffJavaSource> version1ast = input.getModifiableVersion1AST();
		Map<String, SemDiffJavaSource> version2ast = input.getModifiableVersion2AST();
		
		Set<DiffCatResult> results = new HashSet<DiffCatResult>();
		
		for (String filePath1 : version1ast.keySet()) {
			for (String filePath2 : version2ast.keySet()) {
				
				if (comparer.sameFilePath(filePath1, filePath2)) {
					
					SemDiffJavaSource source1 = version1ast.get(filePath1);
					SemDiffJavaSource source2 = version2ast.get(filePath2);
					
					CompilationUnit compilationUnit1 = source1.getCompilationUnit();
					CompilationUnit compilationUnit2 = source2.getCompilationUnit();
					
					Set<DiffCatResult> tempResults = changeDistiller.distill(compilationUnit1, compilationUnit2);
					Set<DiffCatResult> pathed = SetUtil.apply(tempResults, new SetPathsOp(filePath1, filePath2));
					results.addAll(pathed);
				}
			}
		}
		
		return results;
	}
	
	private Set<DiffCatResult> getAddedAndDeletedClasses (DetectorInput input) 
	{
		Map<String, SemDiffJavaSource> version1ast = input.getModifiableVersion1AST();
		Map<String, SemDiffJavaSource> version2ast = input.getModifiableVersion2AST();
		
		Set<DiffCatResult> results = new HashSet<DiffCatResult>();
		
		for (String filePath1 : version1ast.keySet()) {
			
			boolean hasPartner = false;
			for (String filePath2 : version2ast.keySet()) {
				hasPartner |= comparer.sameFilePath(filePath1, filePath2);
			}
			
			if (! hasPartner) {
				Set<ASTNode> delTypes = ASTUtil.getDeclaredTypes(version1ast.get(filePath1).getCompilationUnit());
				
				for (ASTNode delType : delTypes) {
					
					DiffCatResult delResult = new DiffCatResult();
					delResult.setChangeType(ChangeOperation.CLASS_DELETE.name());
					delResult.setNodeFrom(delType);
					delResult.setLeftParent(delType);
					delResult.setPathFrom(filePath1);
					
					results.add(delResult);
				}
			}
		}
		
		for (String filePath2 : version2ast.keySet()) {
			
			boolean hasPartner = false;
			for (String filePath1 : version1ast.keySet()) {
				hasPartner |= comparer.sameFilePath(filePath1, filePath2);
			}
			
			if (! hasPartner) {
				Set<ASTNode> insTypes = ASTUtil.getDeclaredTypes(version2ast.get(filePath2).getCompilationUnit());
				
				for (ASTNode insType : insTypes) {
					
					DiffCatResult insResult = new DiffCatResult();
					insResult.setChangeType(ChangeOperation.CLASS_INSERT.name());
					insResult.setNodeTo(insType);
					insResult.setRightParent(insType);
					insResult.setPathTo(filePath2);
					
					results.add(insResult);
				}
			}
			
		}
		
		return results;
	}
	
	private void findClassRenames (Set<DiffCatResult> changes, Set<DiffCatResult> insDels) {
		
		Set<DiffCatResult> classDels = SetUtil.filterAndCopy(insDels, new IsClassDeleteFilter());
		Set<DiffCatResult> classInss = SetUtil.filterAndCopy(insDels, new IsClassInsertFilter());
		
		CollectionMap<DiffCatResult, DiffCatResult> matchGraph = new CollectionMap<DiffCatResult, DiffCatResult>();
		
		// build K2 match graph
		for (DiffCatResult classDel : classDels) {
			for (DiffCatResult classIns : classInss) {
				matchGraph.add(classDel, classIns);
			}
		}
		
		// trim candidates to find definite renames
		classRenameDetector.getScorer().trimCandidateMap(matchGraph, changes);
		
		// keep those class deletes that are not matched to at least one insert
		for (DiffCatResult classDel : classDels) {
			if (! matchGraph.keySet().contains(classDel)) {
				changes.add(classDel);
			}
		}
		
		// keep those class inserts that are not matched to at least one delete
		for (DiffCatResult classIns : classInss) {
			if (! matchGraph.values().contains(classIns)) {
				changes.add(classIns);
			}
		}
			
		// all others are now lined up as renames
		for (DiffCatResult classDel : matchGraph.keySet()) {
			Collection<DiffCatResult> inserts = matchGraph.get(classDel);	// only one in this
			for (DiffCatResult classIns : inserts) {
				insDels.add( makeRename (classDel, classIns));
				insDels.remove(classDel);
				insDels.remove(classIns);
			}
		}
	}
	
	private DiffCatResult makeRename (DiffCatResult classDel, DiffCatResult classIns) {
		
		DiffCatResult rename = new DiffCatResult();
		
		AbstractTypeDeclaration del = (AbstractTypeDeclaration) classDel.getNodeFrom();
		AbstractTypeDeclaration ins = (AbstractTypeDeclaration) classIns.getNodeTo();
		
		SimpleName delName = del.getName();
		SimpleName insName = ins.getName();
		
		rename.setChangeType(ChangeOperation.CLASS_NAME_UPDATE.name());
		rename.setNodeFrom(delName);
		rename.setNodeTo(insName);
		rename.setLeftParent(del);
		rename.setRightParent(ins);
		rename.setPathFrom(classDel.getPathFrom());
		rename.setPathTo(classIns.getPathTo());
		
		return rename;
	}
	
	private Set<DiffCatResult> getChangesForRenamedFiles (Set<DiffCatResult> classRenames) {
		
		Set<DiffCatResult> changes = new HashSet<DiffCatResult>();
		
		for (DiffCatResult classRename : classRenames) {
			
			ASTNode leftParent = classRename.getLeftParent();
			ASTNode rightParent = classRename.getRightParent();
			
			// get CD changes
			Set<DiffCatResult> results = changeDistiller.distill(leftParent, rightParent);
			
			// make (sometimes artificial) constructor rename events
			getElementSignatureUpdates (classRename, results);
			
			// make sure to add the file paths for viewing 
			changes.addAll (SetUtil.apply(results, new SetPathsOp(classRename.getPathFrom(), classRename.getPathTo())));
		}
		
		return changes;
	}
	
	private void getElementSignatureUpdates (DiffCatResult classRename, Set<DiffCatResult> results) {
		
		ASTNode leftParent = classRename.getLeftParent();
		ASTNode rightParent = classRename.getRightParent();
		
		List<BodyDeclaration> leftFieldsAndMethods = ASTUtil.getFieldsAndMethods((AbstractTypeDeclaration) leftParent);
		List<BodyDeclaration> rightFieldsAndMethods = ASTUtil.getFieldsAndMethods((AbstractTypeDeclaration) rightParent);
		
		Set<DiffCatResult> tempResults = new HashSet<DiffCatResult>();
		
		// Fields and Methods with differences between them
		for (DiffCatResult r : results) {
			
			ASTNode leftParent2 = r.getLeftParent();
			ASTNode rightParent2 = r.getRightParent();
			
			int leftIndex = ListUtil.firstIndexOf(leftFieldsAndMethods, new BasicEqualityFilter<ASTNode>(leftParent2));
			int rightIndex = ListUtil.firstIndexOf(rightFieldsAndMethods, new BasicEqualityFilter<ASTNode>(rightParent2));
			
			if (leftIndex == -1 || rightIndex == -1) {
				continue;
			}
			
			BodyDeclaration leftC = leftFieldsAndMethods.get(leftIndex);
			BodyDeclaration rightC = rightFieldsAndMethods.get(rightIndex);
			
			leftFieldsAndMethods.remove(leftIndex);
			rightFieldsAndMethods.remove(rightIndex);
			
			DiffCatResult sigUpdateResult = makeSigUpdateResult(leftC, classRename.getPathFrom(), rightC, classRename.getPathTo());
			
			tempResults.add(sigUpdateResult);
		}
		
		// The remaining methods had no differences at all
		// Hence we assume their signatures still match exactly
		class ParameterMatcher implements IFilter<BodyDeclaration> {
			
			private final BodyDeclaration toMatch;
			
			public ParameterMatcher (BodyDeclaration toMatch) {
				this.toMatch = toMatch;
			}
			
			public boolean passesFilter(BodyDeclaration t) {
				
				if (toMatch.getNodeType() != t.getNodeType()) {
					return false;
				}
				
				if (toMatch instanceof MethodDeclaration) {
					return methodPassesFilter ((MethodDeclaration) t);
				}
				
				else {
					return fieldPassesFilter ((FieldDeclaration) t);
				}
			}
			
			private boolean fieldPassesFilter(FieldDeclaration t) {
				
				FieldDeclaration toMatch = (FieldDeclaration) this.toMatch;
				
				String fieldNameMatch = ASTUtil.getFieldDeclName(toMatch);
				String fieldNameT = ASTUtil.getFieldDeclName(t);
				
				return fieldNameMatch.equals(fieldNameT);
			}

			public boolean methodPassesFilter (MethodDeclaration t)
			{
				MethodDeclaration toMatch = (MethodDeclaration) this.toMatch;
				
				IMethodBinding matchBinding = toMatch.resolveBinding();
				IMethodBinding tBinding = t.resolveBinding();
				
				ITypeBinding[] paramsToMatch = matchBinding.getParameterTypes();
				ITypeBinding[] params = tBinding.getParameterTypes();
				
				if (paramsToMatch.length != params.length) {
					return false;
				}
				
				for (int i = 0; i < paramsToMatch.length; i++) {
					
					ITypeBinding paramToMatch = paramsToMatch[i];
					ITypeBinding param = params[i];
					
					if (! paramToMatch.getName().equals(param.getName())) {
						return false;
					}
				}
				
				if (matchBinding.isConstructor() != tBinding.isConstructor()) {
					return false;
				}
					
				if (! matchBinding.isConstructor()) {
					return matchBinding.getName().equals(tBinding.getName());
				}
				
				return true;
			}
		}
		
		// we add all signature-matching elements 
		for (int i = 0; i < leftFieldsAndMethods.size(); i++) {
			
			int j = ListUtil.firstIndexOf(rightFieldsAndMethods, new ParameterMatcher(leftFieldsAndMethods.get(i)));
			
			if (j == -1) {
				continue;
			}
			
			tempResults.add(makeSigUpdateResult(leftFieldsAndMethods.get(i), classRename.getPathFrom(), rightFieldsAndMethods.get(j), classRename.getPathTo()));
			
			rightFieldsAndMethods.remove(j);
		}
		
		results.addAll(tempResults);
	}
	
	private DiffCatResult makeSigUpdateResult (BodyDeclaration left, String leftPath, BodyDeclaration right, String rightPath) {
		
		DiffCatResult result = new DiffCatResult();
		
		result.setLeftParent(left);
		result.setRightParent(right);
		result.setPathFrom(leftPath);
		result.setPathTo(rightPath);
		
		setSigUpdateOp (result, left, right);
		
		return result;
	}
	
	private void setSigUpdateOp (DiffCatResult result, BodyDeclaration left, BodyDeclaration right) {
		if (left instanceof MethodDeclaration) {
			MethodDeclaration left2 = (MethodDeclaration) left;
			IMethodBinding bindingL = left2.resolveBinding();
			MethodDeclaration right2 = (MethodDeclaration) right;
			IMethodBinding bindingR = right2.resolveBinding();
			
			if (bindingL.getName().equals(bindingR.getName())) {
				result.setChangeType(ChangeOperation.METHOD_SIGNATURE_UPDATE.name());
			}
			
			else {
				result.setChangeType(ChangeOperation.METHOD_NAME_UPDATE.name());
			}
			
			result.setNodeFrom(left2.getName());
			result.setNodeTo(right2.getName());
		}
		
		else {
			
			FieldDeclaration left2 = (FieldDeclaration) left;
			FieldDeclaration right2 = (FieldDeclaration) right;
			
			result.setNodeFrom(ASTUtil.getFieldDeclNameNode(left2));
			result.setNodeTo(ASTUtil.getFieldDeclNameNode(right2));
			
			result.setChangeType(ChangeOperation.ATTRIBUTE_SIGNATURE_UPDATE.name());
		}
	}
	
	private void addDiffCatDetectedRenames (Set<DiffCatResult> results) {
		Set<DiffCatResult> refined = refine (results, new VariableDeclarationRefiner());
		refined = new FieldRenameReconciler().reconcile(refined);
		refined = new ParamRenameReconciler().reconcile(refined);
		results.clear();
		results.addAll(refined);
	}
	
	private Set<DiffCatResult> refine (Set<DiffCatResult> results, IDiffRefiner refiner) {
		Set<DiffCatResult> newResults = new HashSet<DiffCatResult>();
		for (DiffCatResult result : results) {
			Set<DiffCatResult> refined = refiner.refine(result);
			newResults.addAll (refined);
		}
		
		return newResults;
	}
	
	private Set<DiffCatResult> findRenameAffectedChanges (Set<DiffCatResult> changes) {
		
		// check all type updates -> might just be renamed classes
		changes = new TypeRenameReconciler().reconcile(changes);
		
		// check statements for rename-induced diffs
		changes = new RenameReconciler().reconcile(changes);
		
		// reconcile updates to parent interfaces, parent classes, and declared exceptions 
		changes = new ParentInterfaceInsDelReconciler().reconcile(changes);
		changes = new ParentClassInsDelReconciler().reconcile(changes);
		changes = new ThrownExceptionsReconciler().reconcile(changes);
		
		
		return changes;
	}
	
	private void undoRenames (Set<DiffCatResult> renames, DetectorInput input)
	{
		Map<String, SemDiffJavaSource> version2ast = input.getModifiableVersion2AST();
		
		for (SemDiffJavaSource file : version2ast.values()) {
			CompilationUnit unit = file.getCompilationUnit();
			renameHelper.undoRenames(unit, renames);
		}
	}
	
	private void redoRenames (Set<DiffCatResult> renames, DetectorInput input)
	{
		Map<String, SemDiffJavaSource> modifiableVersion2AST = input.getModifiableVersion2AST();
		
		for (SemDiffJavaSource s : modifiableVersion2AST.values()) {
			
			ASTNode node = s.getCompilationUnit();
			renameHelper.redoRenames(node, renames);
		}
	}
	
	private boolean processedInPhase2 (DiffCatResult phase1Result, Set<DiffCatResult> phase2Results) {
		boolean did = false;
		if (phase1Result.getChangeType().equals(ChangeOperation.AFFECTED_ONLY_BY_RENAMES.name())){
			return false;
		}
		for (DiffCatResult phase2Result : phase2Results) {
			if (did) { continue; }
			did = isPhase2Equivalent(phase2Result, phase1Result);
		}
		
		if (did) {
			return did;
		}
		
		return did;
	}
	
	private boolean isPhase2Equivalent (DiffCatResult phase2Result, DiffCatResult phase1Result) {
		if (phase2Result.getNodeFrom() != null &&
			phase2Result.getNodeFrom() == phase1Result.getNodeFrom()) {
			return true; 
		}
		
		if (phase2Result.getNodeTo() != null && phase1Result.getNodeTo() == phase2Result.getNodeTo()) {
			return true;
		}
		
		return false;
	}
	
	private Set<DiffCatResult> doLocalVarExtraction (Set<DiffCatResult> results) {
		return new LocalVarExtractionReconciler().reconcile(results);
	}
	
	
	private Set<DiffCatResult> refineAndReconcileDiffs (Set<DiffCatResult> results)
	{
		// find trivial super( ) invocations
		results = refine(results, new SuperConstructorRefiner());
		
		// find trivial type updates (fqn -> simple name, vice versa)
		results = refine(results, new TypeUpdateRefiner());
		
		// find trivial string split updates
		results = refine(results, new StringSplitArgRefiner());
		
		// find trivial return; statements
		results = refine(results, new ReturnStatementRefiner());
		
		// find trivial if-statement updates
		results = refine(results, new IfConditionUpdateRefiner());
		
		// find trivial constructor inserts/deletes
		results = refine(results, new ConstructorInsDelReconciler());
		
		// find redundant class cast inserts/deletes
		results = refine(results, new CastExpressionRefiner());
		
		// find trivial assignment inserts/deletes
		results = refine(results, new InitRefiner());
		
		// check for trivial this keyword insertions and deletions
		results = refine(results, new ThisExpressionRefiner());
		
		// check for 'h' to "h" conversions and back
//		results = refine(results, new StringCharInterchangeRefiner());
		
		// check again for type updates induced by renames
		results = new TypeRenameReconciler().reconcile(results);
		
		// check again for rename in statements
		results = new RenameReconciler().reconcile(results);
		
		// fix class-inses-dels that are left over somehow
		results = new ClassInsDelReconciler().reconcile(results);
		
		// remove spurious var extraction refactorings
		results = new VarExtractionReconciler().reconcile(results);
		
		// identify spurious del/ins in small methods
		results = new SmallMethodReconciler().reconcile(results);
		
		return results;
	}
}
