/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.analysis;

import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;

import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;

public class DefaultClassRenameDetector extends AbstractSimilarityScorer {

	final static double MIN_THRESH = 0.5; 

	public DefaultClassRenameDetector () {
		setMinThresh(MIN_THRESH);
	}
	
	@Override
	public double getSimilarityScore (DiffCatResult delete, DiffCatResult insert, Set<DiffCatResult> knowns) {
		
		if (rejectViaHeuristics(delete, insert)) {
			return -1.0;
		}
		
		double elementOverlap = 2 * getElementSignatureOverlap (delete, insert, knowns);
		double size = 0;
		size += getElementSize((AbstractTypeDeclaration) delete.getNodeFrom());
		size += getElementSize((AbstractTypeDeclaration) insert.getNodeTo());
		
		if (size == 0) {
			return -1.0;
		}
		return elementOverlap / size;
		
	}
	
	public boolean rejectViaHeuristics (DiffCatResult delete, DiffCatResult insert) 
	{
		AbstractTypeDeclaration deleted = (AbstractTypeDeclaration) delete.getNodeFrom();
		AbstractTypeDeclaration inserted = (AbstractTypeDeclaration) insert.getNodeTo();
		
		ITypeBinding delBinding = deleted.resolveBinding();
		ITypeBinding insBinding = inserted.resolveBinding();
		
		if (delBinding.isInterface() != insBinding.isInterface()) {
			return true;
		}
		
		return false;
	}
	
	public int getElementSignatureOverlap (DiffCatResult delete, DiffCatResult insert, Set<DiffCatResult> knowns) {
		AbstractTypeDeclaration deleted = (AbstractTypeDeclaration) delete.getNodeFrom();
		AbstractTypeDeclaration nodeTo = (AbstractTypeDeclaration) insert.getNodeTo();
		
		return getElementSignatureOverlap(deleted, nodeTo, knowns);
	}
	
	@SuppressWarnings("rawtypes")
	public int getElementSignatureOverlap (AbstractTypeDeclaration left, AbstractTypeDeclaration right, Set<DiffCatResult> knowns) 
	{
		int signatureOverlap = 0;
		
		List leftDeclarations = left.bodyDeclarations();
		List rightDeclarations = right.bodyDeclarations();
		
		for (Object o : leftDeclarations) {
			
			if (o instanceof FieldDeclaration) {
				if (hasMatchingFieldDecl((FieldDeclaration) o, rightDeclarations, knowns)) {
					signatureOverlap ++;
				}
			}
			
			if (o instanceof MethodDeclaration) {
				if (hasMatchingMethodDecl((MethodDeclaration) o, rightDeclarations, knowns)) {
					signatureOverlap ++;
				}
			}
		}
		
		signatureOverlap += (left.getName().getIdentifier().equals(right.getName().getIdentifier())) ? 1 : 0;
		
		return signatureOverlap;
	}
	
	@SuppressWarnings("rawtypes")
	public int getElementSize (AbstractTypeDeclaration decl) {
		List decls = decl.bodyDeclarations();
		int size = 0;
		for (Object o : decls) {
			size += (o instanceof FieldDeclaration) || (o instanceof MethodDeclaration) ? 1 : 0; 
		}
		return size + 1;	// +1 for the type's name
	}
	
	@SuppressWarnings("rawtypes")
	private boolean hasMatchingFieldDecl (FieldDeclaration declaration, List declarations, Set<DiffCatResult> knowns)
	{
		for (int i = 0; i < declarations.size(); i++) {
			Object o = declarations.get(i);
			
			if (o instanceof FieldDeclaration) {
				if (exactMatches (declaration, (FieldDeclaration) o, knowns)) {
					return true;
				}
			}
		}
		
		return false;
	}
	
	@SuppressWarnings("rawtypes")
	private boolean exactMatches (FieldDeclaration declarationOld, FieldDeclaration declarationNew, Set<DiffCatResult> knowns) {
		
		Type typeOld = declarationOld.getType();
		Type typeNew = declarationNew.getType();
		
		ITypeBinding oldBinding = typeOld.resolveBinding();
		ITypeBinding newBinding = typeNew.resolveBinding();
		
		if (! oldBinding.getName().equals(newBinding.getName())) {
			return false;
		}
		
		List oldFragments = declarationOld.fragments();
		List newFragments = declarationNew.fragments();
		
		if (oldFragments.size() != newFragments.size()) {
			return false;	// TODO handle this better
		}
		
		for (int i = 0; i < oldFragments.size(); i++) {
			
			VariableDeclarationFragment oldFragment = (VariableDeclarationFragment) oldFragments.get(i);
			VariableDeclarationFragment newFragment = (VariableDeclarationFragment) newFragments.get(i);
			
			if (! oldFragment.getName().getIdentifier().equals(newFragment.getName().getIdentifier())) {
				return false;
			}
		}
		
		return true;
	}
	
	@SuppressWarnings("rawtypes")
	private boolean hasMatchingMethodDecl (MethodDeclaration declaration, List declarations, Set<DiffCatResult> knowns)
	{
		for (int i = 0; i < declarations.size(); i++) {
			
			Object o = declarations.get(i);
			
			if (o instanceof MethodDeclaration) {
				if (exactMatches(declaration, (MethodDeclaration) o, knowns)) {
					return true;
				}
			}
			
		}
		
		return false;
	}
	
	private boolean exactMatches (MethodDeclaration declarationOld, MethodDeclaration declarationNew, Set<DiffCatResult> knowns) {
		
		IMethodBinding oldBinding = declarationOld.resolveBinding();
		IMethodBinding newBinding = declarationNew.resolveBinding();
		
		if (! oldBinding.getName().equals(newBinding.getName()) && (! oldBinding.isConstructor() || ! newBinding.isConstructor())) {
			return false;
		}
		
		ITypeBinding[] oldParams = oldBinding.getParameterTypes();
		ITypeBinding[] newParams = newBinding.getParameterTypes();
		
		if (oldParams.length != newParams.length) {
			return false;
		}
		
		for (int i = 0; i < oldParams.length; i++) {
			
			ITypeBinding oldParam = oldParams[i];
			ITypeBinding newParam = newParams[i];
			
			if (! oldParam.getName().equals(newParam.getName())) {
				return false;
			}
			
		}
		
		return true;
	}
}
