/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.cd;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.evolizer.changedistiller.model.classifiers.ChangeType;
import org.evolizer.changedistiller.model.classifiers.EntityType;
import org.evolizer.changedistiller.model.classifiers.SourceRange;
import org.evolizer.changedistiller.model.entities.Delete;
import org.evolizer.changedistiller.model.entities.Insert;
import org.evolizer.changedistiller.model.entities.Move;
import org.evolizer.changedistiller.model.entities.SourceCodeChange;
import org.evolizer.changedistiller.model.entities.SourceCodeEntity;
import org.evolizer.changedistiller.model.entities.StructureEntityVersion;
import org.evolizer.changedistiller.model.entities.Update;

import ca.mcgill.cs.swevo.diffcat.cd.finder.ANodeFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.BreakStatementFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.CDUtil;
import ca.mcgill.cs.swevo.diffcat.cd.finder.CatchClauseFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.ConditionFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.ContinueStatementFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.ElseBlockFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.FieldDeclarationFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.ForStatementFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.HeuristicMethodInsertFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.LabelledStatementFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.MethodDeclarationFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.ReturnStatementFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.SoftMethodDeclarationFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.StatementFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.TryStatementFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.TypeDeclarationFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.UltraFieldDeclarationFinder;
import ca.mcgill.cs.swevo.diffcat.cd.finder.UltraMethodDeclarationFinder;
import ca.mcgill.cs.swevo.diffcat.model.ChangeOp;
import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.util.CollectionMap;

/**
 * @author david
 */
public class CDDiffMapper {

	private final Logger log = Logger.getLogger(getClass().getName());
	
	/**
	 * Hashes right renamed field root entities to their associated left field declarations
	 */
	private Map<String, FieldDeclaration> fieldSignatureUpdates;
	
	/**
	 * Hashes right updated method signature root entities to their associated left method declarations 
	 */
	private Map<String, MethodDeclaration> methodSignatureUpdates;
	
	/**
	 * Hashes right renamed type root entities to their associated left type declarations 
	 */
	private Map<String, AbstractTypeDeclaration> classRenamesBackward;
	
	/**
	 * The left AST from which CD distilled differences
	 */
	private ASTNode leftType;
	
	/**
	 * The right AST from which CD distilled differences
	 */
	private ASTNode rightType;
	
	
	public CDDiffMapper () 
	{
		fieldSignatureUpdates = new HashMap<String, FieldDeclaration>();
		methodSignatureUpdates = new HashMap<String, MethodDeclaration> ();
		classRenamesBackward = new HashMap<String, AbstractTypeDeclaration>();
	}
	
	public void setLeftParent (ASTNode leftType) {
		this.leftType = leftType;
	}
	
	public void setRightParent (ASTNode rightType) {
		this.rightType = rightType;
	}
	
	public Set<DiffCatResult> mapCDResults (Collection<SourceCodeChange> cdDiffs) {
		
		initFieldSignatureUpdates(cdDiffs);		// to resolve multi-signature field declaration changes later on
		initMethodSignatureUpdates (cdDiffs); 	// to resolve multi-signature method declaration changes later on
		initClassRenames(cdDiffs);				// to resolve multi-signature type declaration changes later on 	
		
		Set<DiffCatResult> results = new HashSet<DiffCatResult>();
		
		for (SourceCodeChange cdDiff : cdDiffs) {
			try {
				DiffCatResult mapped = mapCDChange(cdDiff);
				if (mapped == null) {
					log.log(Level.SEVERE, "Could not map cdDiff: " + cdDiff);
//					mapCDChange(cdDiff);
					continue;
				}
				
				results.add(mapped);
				
				} 
			catch (NullPointerException e)  {
					log.log(Level.SEVERE, "Could not map cdDiff: " + cdDiff);
				}
		
		}
		
		return results;
	}
	
	
	public void initFieldSignatureUpdates (Collection<SourceCodeChange> cdDiffs) {
		fieldSignatureUpdates.clear();
		
		CollectionMap<StructureEntityVersion, SourceCodeChange> sigs = groupByFieldSigs(cdDiffs);
		
		for (StructureEntityVersion root : sigs.keySet())
		{
			Collection<SourceCodeChange> changes = sigs.get(root);
			
			FieldDeclaration declaration = getRightFieldDeclaration(root);
			declaration = getRightFieldDeclaration(root);
			UltraFieldDeclarationFinder finder = new UltraFieldDeclarationFinder(declaration, changes);
			FieldDeclaration leftField = (FieldDeclaration) discover(leftType, finder);
			
			fieldSignatureUpdates.put(root.toString(), leftField);
		}
	}

	public void initMethodSignatureUpdates( Collection<SourceCodeChange> changeSet) {
		CollectionMap<StructureEntityVersion,SourceCodeChange> sigs = groupByMethodSigs(changeSet);
		
		for (StructureEntityVersion root : sigs.keySet()) {
			Collection<SourceCodeChange> changes = sigs.get(root); 
			
			MethodDeclaration declaration = getRightMethodDeclaration(root);
			if (declaration == null) {
				log.log(Level.SEVERE, "Could not find right method declaration: " + root);
				continue;
			}
			UltraMethodDeclarationFinder finder2 = new UltraMethodDeclarationFinder(declaration, changes);
			MethodDeclaration declaration2 = (MethodDeclaration) discover(leftType, finder2);
			
			methodSignatureUpdates.put(root.toString(), declaration2);
		}
	}
	
	private CollectionMap<StructureEntityVersion, SourceCodeChange> groupByFieldSigs (Collection<SourceCodeChange> diffs)
	{
		CollectionMap<StructureEntityVersion, SourceCodeChange> map = new CollectionMap<StructureEntityVersion, SourceCodeChange>();
		
		for (SourceCodeChange change : diffs) {
			if (change.getChangeType().equals(ChangeType.ATTRIBUTE_RENAMING)) {
				map.add(change.getRootEntity(), change);
			}
			else if (change.getChangeType().equals(ChangeType.ATTRIBUTE_TYPE_CHANGE)) {
				map.add(change.getRootEntity(), change);
			}
		}
		
		return map;
	}
	
	private CollectionMap<StructureEntityVersion, SourceCodeChange> groupByMethodSigs (Collection<SourceCodeChange> diffs)
	{
		CollectionMap<StructureEntityVersion, SourceCodeChange> map = new CollectionMap<StructureEntityVersion, SourceCodeChange>();
		
		for (SourceCodeChange change : diffs) {
			if (change.getChangeType().equals(ChangeType.METHOD_RENAMING)) {
				map.add(change.getRootEntity(), change);
			}
			else if (change.getChangeType().equals(ChangeType.PARAMETER_DELETE)) {
				map.add(change.getRootEntity(), change);
			}
			else if (change.getChangeType().equals(ChangeType.PARAMETER_INSERT)) {
				map.add(change.getRootEntity(), change);
			}
			else if (change.getChangeType().equals(ChangeType.PARAMETER_ORDERING_CHANGE)) {
				map.add(change.getRootEntity(), change);
			}
			else if (change.getChangeType().equals(ChangeType.PARAMETER_TYPE_CHANGE)) {
				map.add(change.getRootEntity(), change);
			}
			else if (change.getChangeType().equals(ChangeType.PARAMETER_RENAMING)) {
				map.add(change.getRootEntity(), change);
			}
		}
		
		return map;
	}

	public void initClassRenames(Collection<SourceCodeChange> changeSet) {
		for (SourceCodeChange change : changeSet) {
			if (change.getChangeType().equals(ChangeType.CLASS_RENAMING)) {
				try {
					DiffCatResult r = mapClassRenameChange((Update) change);
					addClassRename (change.getRootEntity(), r);
				} catch (Exception e) {
					log.severe("Error! Change Distiller result not handled properly: " + change.getLabel());
				}
			}
		}
	}
	
	public void addClassRename (StructureEntityVersion root, DiffCatResult r) {
		String rightRootString = root.toString();
		classRenamesBackward.put(rightRootString, (AbstractTypeDeclaration) r.getLeftParent());
	}
	
	public Map<String, FieldDeclaration> getFieldSignatureUpdates() {
		return fieldSignatureUpdates;
	}

	public Map<String, MethodDeclaration> getMethodSignatureUpdates() {
		return methodSignatureUpdates;
	}

	public Map<String, AbstractTypeDeclaration> getClassRenames() {
		return classRenamesBackward;
	}

	private DiffCatResult mapCDChange ( SourceCodeChange change)
	{
		ChangeType changeType = change.getChangeType();
		
		DiffCatResult result = null;
		if (changeType == ChangeType.ADDING_ATTRIBUTE_MODIFIABILITY) {
			result = mapAddingAttributeModifiabilityChange ((Delete) change);
		} else if (changeType == ChangeType.ADDING_CLASS_DERIVABILITY) {
			result = mapAddingClassDerivabilityChange ((Delete) change);
		} else if (changeType == ChangeType.ADDING_METHOD_OVERRIDABILITY) {
			result = mapAddingMethodOverridabilityChange ((Delete) change);
		} else if (changeType == ChangeType.ADDITIONAL_CLASS) {
			result = mapAdditionalClassChange ((Insert) change);
		} else if (changeType == ChangeType.ADDITIONAL_FUNCTIONALITY) {
			result = mapAdditnlFuncltyChange((Insert) change);
		} else if (changeType == ChangeType.ADDITIONAL_OBJECT_STATE) {
			result = mapAdditionalObjectStateChange ((Insert) change);
		} else if (changeType == ChangeType.ALTERNATIVE_PART_DELETE) {
			result = mapAlternativePartDeleteChange ((Delete) change);
		} else if (changeType == ChangeType.ALTERNATIVE_PART_INSERT) {
			result = mapAlternativePartInsertChange ((Insert) change);
		} else if (changeType == ChangeType.ATTRIBUTE_RENAMING) {
			result = mapAttributeRenameChange ((Update) change);
		} else if (changeType == ChangeType.ATTRIBUTE_TYPE_CHANGE) {
			result = mapAttributeTypeChange ((Update) change);
		} else if (changeType == ChangeType.CLASS_RENAMING) {
			result = mapClassRenameChange ((Update) change);
		} else if (changeType == ChangeType.COMMENT_INSERT) {
			result = mapCommentInsertChange((Insert) change);
		} else if (changeType == ChangeType.COMMENT_DELETE) {
			result = mapCommentDeleteChange((Delete) change);
		} else if (changeType == ChangeType.COMMENT_MOVE) {
			result = mapCommentMoveChange((Move) change);
		} else if (changeType == ChangeType.COMMENT_UPDATE) {
			result = mapCommentUpdateChange ((Update) change);
		} else if (changeType == ChangeType.CONDITION_EXPRESSION_CHANGE) {
			result = mapConditionExpressionChange ((Update) change);
		} else if (changeType == ChangeType.DECREASING_ACCESSIBILITY_CHANGE) {
			result = mapAccessibilityChange (change);
		} else if (changeType == ChangeType.DOC_UPDATE) {
			result = mapJavadocUpdateChange((Update) change);
		} else if (changeType == ChangeType.DOC_INSERT) {
			result = mapJavadocInsertChange((Insert) change);
		} else if (changeType == ChangeType.DOC_DELETE) {
			result = mapJavadocDeleteChange((Delete) change);
		} else if (changeType == ChangeType.INCREASING_ACCESSIBILITY_CHANGE) {
			result = mapAccessibilityChange (change);
		} else if (changeType == ChangeType.METHOD_RENAMING) {
			result = mapMethodRenameChange ((Update) change);
		} else if (changeType == ChangeType.PARAMETER_TYPE_CHANGE) {
			result = mapParamTypeChange((Update) change); 
		} else if (changeType == ChangeType.PARAMETER_DELETE) {
			result = mapParamDeleteChange ((Delete) change);
		} else if (changeType == ChangeType.PARAMETER_RENAMING) {
			result = mapParamRenameChange ((Update) change);
		} else if (changeType == ChangeType.PARAMETER_INSERT) {
			result = mapParamInsertChange ((Insert) change);
		} else if (changeType == ChangeType.PARAMETER_ORDERING_CHANGE) {
			result = mapParamOrderingChange ((Move) change);
		} else if (changeType == ChangeType.PARENT_CLASS_INSERT) {
			result = mapParentClassInsertChange ((Insert) change);
		} else if (changeType == ChangeType.PARENT_CLASS_CHANGE) {
			result = mapParentClassUpdateChange ((Update) change);
		} else if (changeType == ChangeType.PARENT_CLASS_DELETE) {
			result = mapParentClassDeleteChange ((Delete) change);
		} else if (changeType == ChangeType.PARENT_INTERFACE_CHANGE) {
			result = mapParentInterfaceUpdateChange ((Update) change);
		} else if (changeType == ChangeType.PARENT_INTERFACE_INSERT) {
			result = mapParentInterfaceInsertChange ((Insert) change);
		} else if (changeType == ChangeType.PARENT_INTERFACE_DELETE) {
			result = mapParentInterfaceDeleteChange ((Delete) change);
		} else if (changeType == ChangeType.REMOVED_CLASS) {
			result = mapRemovedClassChange ((Delete) change);
		} else if (changeType == ChangeType.REMOVED_FUNCTIONALITY) {
			result = mapRemovedFunctlyChange((Delete) change);
		} else if (changeType == ChangeType.REMOVED_OBJECT_STATE) {
			result = mapRemovedObjectStateChange ((Delete) change);
		} else if (changeType == ChangeType.REMOVING_CLASS_DERIVABILITY) {
			result = mapRemovingClassDerivabilityChange ((Insert) change);
		} else if (changeType == ChangeType.REMOVING_METHOD_OVERRIDABILITY) {
			result = mapRemovingMethodModifiabilityChange ((Insert) change);
		} else if (changeType == ChangeType.REMOVING_ATTRIBUTE_MODIFIABILITY) {
			result = mapRemovingAttributeModifiability((Insert) change);
		} else if (changeType == ChangeType.RETURN_TYPE_DELETE) {
			result = mapReturnTypeDeleteChange ((Delete) change);
		} else if (changeType == ChangeType.RETURN_TYPE_INSERT) {
			result = mapReturnTypeInsertChange ((Insert) change);
		} else if (changeType == ChangeType.RETURN_TYPE_CHANGE) {
			result = mapReturnTypeChange ((Update) change);
		} else if (changeType == ChangeType.STATEMENT_UPDATE) {
			result = mapStatementUpdateChange ((Update) change);
		} else if (changeType == ChangeType.STATEMENT_DELETE) {
			result = mapStatementDeleteChange ((Delete) change);
		} else if (changeType == ChangeType.STATEMENT_ORDERING_CHANGE) {
			result = mapStatementOrderingChange ((Move) change);
		} else if (changeType == ChangeType.STATEMENT_INSERT) {
			result = mapStatementInsertChange ((Insert) change);
		} else if (changeType == ChangeType.STATEMENT_PARENT_CHANGE) {
			result = mapStatementParentChange ((Move) change);
		} else if (changeType == ChangeType.UNCLASSIFIED_CHANGE){
			result = mapUnclassifiedChange (change);
		} else {
			result = null;	// should not happen
		}
		
		return result;
	}

	private DiffCatResult mapUnclassifiedChange(  SourceCodeChange change) {
		// Unclassified changes include: i. un/making something static, ii. synchronized
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		ASTNode leftParent = null;
		ASTNode rightParent = null;
		
		ASTNode nodeFrom = null;
		ASTNode nodeTo = null;
		
		if (rootEntity.getType().equals(EntityType.TYPE_DECLARATION) ||
			rootEntity.getType().equals(EntityType.CLASS)) 
		{
			leftParent = getLeftTypeDeclaration(rootEntity);
			rightParent = getRightTypeDeclaration(rootEntity);
			
			nodeFrom = ((AbstractTypeDeclaration) leftParent).getName();	// to appease the gods
			nodeTo = ((AbstractTypeDeclaration) rightParent).getName();
		}
		
		if (rootEntity.getType().equals(EntityType.FIELD_DECLARATION) || 
			rootEntity.getType().equals(EntityType.ATTRIBUTE))
		{
			leftParent = getLeftFieldDeclaration(rootEntity);
			rightParent = getRightFieldDeclaration(rootEntity);
			
			nodeFrom = ((FieldDeclaration) leftParent).getType();
			nodeTo = ((FieldDeclaration) rightParent).getType();
		}
		
		if (rootEntity.getType().equals(EntityType.METHOD_DECLARATION)||
			rootEntity.getType().equals(EntityType.METHOD))
		{
			leftParent = getLeftMethodDeclaration(rootEntity);
			rightParent = getRightMethodDeclaration(rootEntity);
			
			nodeFrom = ((MethodDeclaration) leftParent).getName();
			nodeTo = ((MethodDeclaration) rightParent).getName();
		}
		
		return makeResult(leftParent, rightParent, nodeFrom, nodeTo, ChangeOperation.UNCLASSIFIED);
	}

	private DiffCatResult mapReturnTypeInsertChange(  Insert change) {
		
		MethodDeclaration leftDeclaration = getLeftMethodDeclaration(change.getRootEntity());
		MethodDeclaration rightDeclaration = getRightMethodDeclaration(change.getRootEntity());
		
		Type leftType = leftDeclaration.getReturnType2();
		Type rightType = rightDeclaration.getReturnType2();
		
		ChangeOperation op = ChangeOperation.RETURN_TYPE_UPDATE;
		
		return makeResult(leftDeclaration, rightDeclaration, leftType, rightType, op);
	}

	private DiffCatResult mapReturnTypeDeleteChange(  Delete change) {
		
		MethodDeclaration leftDeclaration = getLeftMethodDeclaration(change.getRootEntity());
		MethodDeclaration rightDeclaration = getRightMethodDeclaration(change.getRootEntity());
		
		Type leftType = leftDeclaration.getReturnType2();
		Type rightType = rightDeclaration.getReturnType2();
		
		ChangeOperation op = ChangeOperation.RETURN_TYPE_UPDATE;
		
		return makeResult(leftDeclaration, rightDeclaration, leftType, rightType, op);
	}

	private DiffCatResult mapRemovingClassDerivabilityChange(  Insert change) {
		
		AbstractTypeDeclaration leftDeclaration = getLeftTypeDeclaration(change.getRootEntity());
		AbstractTypeDeclaration rightDeclaration = getRightTypeDeclaration(change.getRootEntity());
		
		SimpleName leftName = leftDeclaration.getName();
		
		ChangeOperation op = ChangeOperation.CLASS_DERIVABILITY_DELETE;
		
		return makeResult(leftDeclaration, rightDeclaration, leftName, null, op);
	}

	private DiffCatResult mapAddingClassDerivabilityChange(  Delete change) {
		
		AbstractTypeDeclaration leftDeclaration = getLeftTypeDeclaration(change.getRootEntity());
		AbstractTypeDeclaration rightDeclaration = getRightTypeDeclaration(change.getRootEntity());
		
		SimpleName rightName = rightDeclaration.getName();
		
		ChangeOperation op = ChangeOperation.CLASS_DERIVABILITY_INSERT;
		
		return makeResult(leftDeclaration, rightDeclaration, null, rightName, op);
		
	}

	private DiffCatResult mapAddingAttributeModifiabilityChange(  Delete change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		FieldDeclaration leftParent = getLeftFieldDeclaration(rootEntity);
		
		FieldDeclaration rightParent = getRightFieldDeclaration(rootEntity); 
		VariableDeclarationFragment name = (VariableDeclarationFragment) rightParent.fragments().get(0);
		
		ChangeOperation op = ChangeOperation.ATTRIBUTE_MODIFIABILITY_DELETE;
		return makeResult (leftParent, rightParent, null, name, op);
	}

	private DiffCatResult mapAddingMethodOverridabilityChange( Delete change) {
		
		MethodDeclaration declaration = getLeftMethodDeclaration(change.getRootEntity());
		SimpleName name = declaration.getName();
		ChangeOperation op = ChangeOperation.METHOD_OVERRIDABILITY_MODIFIER_DELETE;
		
		MethodDeclaration right = getRightMethodDeclaration(change.getRootEntity());
		
		return makeResult(declaration, right, name, null, op);
	}

	private DiffCatResult mapParentInterfaceDeleteChange( Delete change) {
		
		AbstractTypeDeclaration leftParent = getLeftTypeDeclaration(change.getRootEntity());
		Type interfazz = retrieveInterface(ASTUtil.getInterfaceTypes(leftParent), change.getChangedEntity());
		
		AbstractTypeDeclaration rightParent = getRightTypeDeclaration(change.getRootEntity());
		
		ChangeOperation op = ChangeOperation.PARENT_INTERFACE_DELETE;
		
		return makeResult(leftParent, rightParent, interfazz, null, op);
	}

	private DiffCatResult mapParentInterfaceInsertChange(  Insert change) {
		
		AbstractTypeDeclaration leftParent = getRightTypeDeclaration(change.getRootEntity());
		
		AbstractTypeDeclaration rightParent = getRightTypeDeclaration(change.getRootEntity());
		Type interfazz = retrieveInterface(ASTUtil.getInterfaceTypes(rightParent), change.getChangedEntity());
		
		ChangeOperation op = ChangeOperation.PARENT_INTERFACE_INSERT; 
		
		return makeResult(leftParent, rightParent, null, interfazz, op);
	}

	private DiffCatResult mapAdditionalClassChange( Insert change) {
		
		TypeDeclarationFinder rightFinder = new TypeDeclarationFinder(change.getChangedEntity());
		ASTNode rightNode = discover(rightType, rightFinder);
		ASTNode rightParent = ASTUtil.getTypeContainer(rightNode.getParent()); // get the declaring type
		
		ChangeOperation op = ChangeOperation.CLASS_INSERT;
		return makeResult(null, rightParent, null, rightNode, op);
	}

	private DiffCatResult mapRemovedClassChange(  Delete change) {
		
		TypeDeclarationFinder leftFinder = new TypeDeclarationFinder(change.getChangedEntity());
		ASTNode leftNode = discover(leftType, leftFinder);
		ASTNode parent = ASTUtil.getTypeContainer(leftNode.getParent()); // get the declaring type
		
		ChangeOperation op = ChangeOperation.CLASS_DELETE;
		return makeResult(parent, null, leftNode, null, op);
	}
	
	private DiffCatResult mapParentInterfaceUpdateChange(  Update change) {
		
		AbstractTypeDeclaration rightNode = getRightTypeDeclaration(change.getRootEntity());
		Type rightType = retrieveInterface(ASTUtil.getInterfaceTypes(rightNode), change.getNewEntity());
		
		AbstractTypeDeclaration leftNode = getLeftTypeDeclaration(change.getRootEntity());
		Type leftType = retrieveInterface(ASTUtil.getInterfaceTypes(leftNode), change.getChangedEntity());
		
		ChangeOperation op = ChangeOperation.PARENT_INTERFACE_UPDATE;
		return makeResult(leftNode, rightNode, leftType, rightType, op);
	}

	@SuppressWarnings("rawtypes")
	private Type retrieveInterface (List interfaces, SourceCodeEntity entity) {
		String name = entity.getUniqueName();
		for (int i = 0; i < interfaces.size();i++) {
			Type type = (Type) interfaces.get(i);
			ITypeBinding binding = type.resolveBinding();
			if (binding.getName().equals(name)) {
				return type;
			}
		}
		return null;
	}
	
	private DiffCatResult mapClassRenameChange(  Update change) {
		
		TypeDeclarationFinder leftFinder = new TypeDeclarationFinder(change.getChangedEntity());
		AbstractTypeDeclaration leftNode = (AbstractTypeDeclaration) discover(leftType, leftFinder);
		SimpleName leftName = leftNode.getName();
		
		TypeDeclarationFinder rightFinder = new TypeDeclarationFinder(change.getNewEntity());
		AbstractTypeDeclaration rightNode = (AbstractTypeDeclaration) discover(rightType, rightFinder);
		SimpleName rightName = rightNode.getName();
		
		ChangeOperation op = ChangeOperation.CLASS_NAME_UPDATE;
		return makeResult (leftNode, rightNode, leftName, rightName, op);
	}

	private DiffCatResult mapAttributeRenameChange(  Update change) {
		
		// assume there is only field name. 
		
		FieldDeclarationFinder rightFinder = new FieldDeclarationFinder(change.getNewEntity());
		FieldDeclaration rightField = (FieldDeclaration) discover (rightType, rightFinder);
		SimpleName rightName = ((VariableDeclarationFragment) rightField.fragments().get(0)).getName();
		
		FieldDeclarationFinder leftFinder = new FieldDeclarationFinder (change.getChangedEntity());
		FieldDeclaration leftField = (FieldDeclaration) discover (leftType, leftFinder);
		SimpleName leftName = ((VariableDeclarationFragment) leftField.fragments().get(0)).getName();
		
		ChangeOperation op = ChangeOperation.ATTRIBUTE_NAME_UPDATE;
		return makeResult (leftField, rightField, leftName, rightName, op);
	}

	private DiffCatResult mapAttributeTypeChange( Update change) {
		
		FieldDeclaration leftField = getLeftFieldDeclaration(change.getRootEntity());
		Type leftType = leftField.getType();
		
		FieldDeclaration rightField = getRightFieldDeclaration(change.getRootEntity());
		Type rightType = rightField.getType();
		
		ChangeOperation op = ChangeOperation.ATTRIBUTE_TYPE_UPDATE;
		return makeResult(leftField, rightField, leftType, rightType, op);
	}

	private DiffCatResult mapRemovingAttributeModifiability( Insert change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		FieldDeclaration leftParent = getLeftFieldDeclaration(rootEntity); 
		VariableDeclarationFragment name = (VariableDeclarationFragment) leftParent.fragments().get(0);
		
		FieldDeclaration rightParent = getRightFieldDeclaration(rootEntity);
		
		ChangeOperation op = ChangeOperation.ATTRIBUTE_MODIFIABILITY_DELETE;
		return makeResult (leftParent, rightParent, name, null, op);
	}
	
	private DiffCatResult mapRemovingMethodModifiabilityChange( Insert change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(rootEntity);
		
		MethodDeclaration rightParent = getRightMethodDeclaration(rootEntity);
		SimpleName name = rightParent.getName();
		
		ChangeOperation op = ChangeOperation.METHOD_OVERRIDABILITY_MODIFIER_INSERT;
		return makeResult(leftParent, rightParent, null, name, op);
	}

	private DiffCatResult mapJavadocInsertChange(  Insert change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		BodyDeclaration leftParent = getLeftBodyDeclaration(rootEntity);
		BodyDeclaration rightParent = getRightBodyDeclaration(rootEntity);
		
		ASTNode rightDoc = rightParent != null? rightParent.getJavadoc() : null;
		
		return makeResult(leftParent, rightParent, null, rightDoc, ChangeOperation.DOC_INSERT);
	}

	private DiffCatResult mapJavadocDeleteChange(  Delete change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		BodyDeclaration leftParent = getLeftBodyDeclaration(rootEntity);
		BodyDeclaration rightParent = getRightBodyDeclaration(rootEntity);
		
		ASTNode leftDoc = leftParent != null? leftParent.getJavadoc() : null;
		
		return makeResult(leftParent, rightParent, leftDoc, null, ChangeOperation.DOC_DELETE);
	}
	
	private DiffCatResult mapAdditionalObjectStateChange( Insert change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		AbstractTypeDeclaration	leftParent = getLeftTypeDeclaration(rootEntity);
		AbstractTypeDeclaration rightParent = getRightTypeDeclaration(rootEntity);
		
		ANodeFinder rightFieldFinder = new FieldDeclarationFinder(change.getChangedEntity());
		ASTNode rightNode = discover (rightParent, rightFieldFinder);
		
		ChangeOperation op = ChangeOperation.ATTRIBUTE_INSERT;
		return makeResult(leftParent, rightParent, null, rightNode, op);
	}
	
	private DiffCatResult mapRemovedObjectStateChange(  Delete change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		AbstractTypeDeclaration leftDeclaration = getLeftTypeDeclaration(rootEntity);
		AbstractTypeDeclaration rightDeclaration = getRightTypeDeclaration(rootEntity);
		
		ANodeFinder leftFieldFinder = new FieldDeclarationFinder(change.getChangedEntity());
		ASTNode leftNode = discover (leftDeclaration, leftFieldFinder);
		
		ChangeOperation op = ChangeOperation.ATTRIBUTE_DELETE;
		return makeResult(leftDeclaration, rightDeclaration, leftNode, null, op);
	}

	private DiffCatResult mapConditionExpressionChange( Update change) {
		
		StructureEntityVersion root = change.getRootEntity();
		SourceCodeEntity parentEntity = change.getParentEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(root);
		MethodDeclaration rightParent = getRightMethodDeclaration(root);
		
		ASTNode leftSearchParent = findParentStatement(parentEntity, leftParent);
		ASTNode leftStatement = findRegularStatement(change, change.getChangedEntity(), leftParent, leftSearchParent);
		
		ASTNode rightSearchParent = findParentStatement(parentEntity, rightParent);
		ASTNode rightStatement = findRegularStatement(change, change.getNewEntity(), rightParent, rightSearchParent);
		
		ChangeOperation op = ChangeOperation.getOperation(ChangeOp.UPDATE, change.getChangedEntity().getType());
		DiffCatResult result = makeResult(leftParent, rightParent, leftStatement, rightStatement, op);
		return result;
	}
	
	private DiffCatResult mapAccessibilityChange(  SourceCodeChange change) {
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		BodyDeclaration leftParent = getLeftBodyDeclaration(rootEntity);
		BodyDeclaration rightParent = getRightBodyDeclaration(rootEntity);
		
		ASTNode leftNode = getAccessibilityModifier(leftParent.modifiers());
		ASTNode rightNode = getAccessibilityModifier(rightParent.modifiers());
		
		ChangeOperation op = null;
		if (leftNode == null) {
			if (leftParent.getNodeType() == ASTNode.FIELD_DECLARATION) {
				op = ChangeOperation.ATTRIBUTE_ACCESSIBILITY_INSERT;
			}
			else if (leftParent.getNodeType() == ASTNode.METHOD_DECLARATION){
				op = ChangeOperation.METHOD_ACCESSIBILITY_INSERT;
			}
			else {
				op = ChangeOperation.TYPE_ACCESSIBILITY_INSERT;
			}
		}
		else if (rightNode == null) {
			if (leftParent.getNodeType() == ASTNode.FIELD_DECLARATION) {
				op = ChangeOperation.ATTRIBUTE_ACCESSIBILITY_DELETE;
			}
			else if (leftParent.getNodeType() == ASTNode.METHOD_DECLARATION){
				op = ChangeOperation.METHOD_ACCESSIBILITY_DELETE;
			}
			else {
				op = ChangeOperation.TYPE_ACCESSIBILITY_DELETE;
			}
		}
		else {
			if (leftParent.getNodeType() == ASTNode.FIELD_DECLARATION) {
				op = ChangeOperation.ATTRIBUTE_ACCESSIBILITY_UPDATE;
			}
			else if (leftParent.getNodeType() == ASTNode.METHOD_DECLARATION){
				op = ChangeOperation.METHOD_ACCESSIBILITY_UPDATE;
			}
			else {
				op = ChangeOperation.TYPE_ACCESSIBILITY_UPDATE;
			}
		}
		
		return makeResult(leftParent, rightParent, leftNode, rightNode, op);
	}
	
	@SuppressWarnings("rawtypes")
	private Modifier getAccessibilityModifier (List modifiers) {
		for (int i = 0; i < modifiers.size(); i++) {
			if (!(modifiers.get(i) instanceof Modifier)) {
				continue;
			}
			Modifier mod = (Modifier) modifiers.get(i);
			if (mod.isPrivate()) {
				return mod;
			}
			if (mod.isPublic()) {
				return mod;
			}
			if (mod.isProtected()) {
				return mod;
			}
		}
		
		return null;
	}
	
	private DiffCatResult mapParentClassUpdateChange(  Update change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		AbstractTypeDeclaration leftTypeDecl = getLeftTypeDeclaration(rootEntity);
		AbstractTypeDeclaration rightTypeDecl = getRightTypeDeclaration(rootEntity);
		
		ASTNode leftNode = ASTUtil.getSuperType(leftTypeDecl);
		ASTNode rightNode = ASTUtil.getSuperType(rightTypeDecl);
		
		ChangeOperation op = ChangeOperation.PARENT_CLASS_UPDATE;
		return makeResult(leftTypeDecl, rightTypeDecl, leftNode, rightNode, op);
	}
	
	private DiffCatResult mapParentClassInsertChange(  Insert change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		AbstractTypeDeclaration leftDecl = getLeftTypeDeclaration(rootEntity);
		AbstractTypeDeclaration rightTypeDecl = getRightTypeDeclaration(rootEntity);
		
		ASTNode rightNode = ASTUtil.getSuperType(rightTypeDecl);
		
		ChangeOperation op = ChangeOperation.PARENT_CLASS_INSERT;
		return makeResult(leftDecl, rightTypeDecl, null, rightNode, op);
	}

	private DiffCatResult mapParentClassDeleteChange( Delete change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		AbstractTypeDeclaration leftTypeDecl = getLeftTypeDeclaration(rootEntity);
		AbstractTypeDeclaration rightTypeDecl = getRightTypeDeclaration(rootEntity);
		ASTNode leftNode = ASTUtil.getSuperType(leftTypeDecl);
		
		ChangeOperation op = ChangeOperation.PARENT_CLASS_DELETE;
		return makeResult (leftTypeDecl, rightTypeDecl, leftNode, null, op);
	}
	
	private DiffCatResult mapStatementParentChange( Move change) {
		return mapStatementOrderingChange(change);
	}

	private DiffCatResult mapStatementOrderingChange( Move change) {
		StructureEntityVersion root = change.getRootEntity();
		SourceCodeEntity leftParentEntity = change.getParentEntity();
		SourceCodeEntity rightParentEntity = change.getNewParentEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(root);	// assumes same method decl
		MethodDeclaration rightParent = getRightMethodDeclaration(root);
		
		ASTNode leftSearchParent = findParentStatement(leftParentEntity, leftParent);
		ASTNode rightSearchParent = findParentStatement(rightParentEntity, rightParent);
		
		ASTNode leftStatement = findRegularStatement(change.getChangedEntity(), leftParent, leftSearchParent);
		ASTNode rightStatement = findRegularStatement(change.getNewEntity(), rightParent, rightSearchParent); 
		
		ChangeOperation op = ChangeOperation.getOperation(ChangeOp.MOVE, change.getNewEntity().getType());
		DiffCatResult result = makeResult(leftParent, rightParent, leftStatement, rightStatement, op);
		return result;
	}

	private DiffCatResult mapStatementDeleteChange( Delete change) {
		
		StructureEntityVersion root = change.getRootEntity();
		SourceCodeEntity parentEntity = change.getParentEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(root);
		MethodDeclaration rightParent = getRightMethodDeclaration(root);
		
		ASTNode searchParent = findParentStatement(parentEntity, leftParent);
		ASTNode leftStatement = findRegularStatement(change.getChangedEntity(), leftParent, searchParent);
		
		ChangeOperation op = ChangeOperation.getOperation(ChangeOp.DELETE, change.getChangedEntity().getType());
		return makeResult(leftParent, rightParent, leftStatement, null, op);
	}
	
	private DiffCatResult mapStatementInsertChange( Insert change) {
		
		StructureEntityVersion root = change.getRootEntity();
		SourceCodeEntity parentEntity = change.getParentEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(root);
		MethodDeclaration rightParent = getRightMethodDeclaration(root);
		
		ASTNode searchParent = findParentStatement(parentEntity, rightParent);
		ASTNode rightStatement = findRegularStatement(change.getChangedEntity(), rightParent, searchParent);
		
		ChangeOperation op = ChangeOperation.getOperation(ChangeOp.INSERT, change.getChangedEntity().getType());
		return makeResult(leftParent, rightParent, null, rightStatement, op);
	}

	private DiffCatResult mapStatementUpdateChange( Update change) {
		
		StructureEntityVersion root = change.getRootEntity();
		SourceCodeEntity parentEntity = change.getParentEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(root);
		MethodDeclaration rightParent = getRightMethodDeclaration(root);
		
		ASTNode leftSearchParent = findParentStatement(parentEntity, leftParent);
		ASTNode leftStatement = findRegularStatement(change.getChangedEntity(), leftParent, leftSearchParent);
		
		ASTNode rightSearchParent = findParentStatement(parentEntity, rightParent);
		ASTNode rightStatement = findRegularStatement(change.getNewEntity(), rightParent, rightSearchParent);
		
		ChangeOperation op = ChangeOperation.getOperation(ChangeOp.UPDATE, change.getChangedEntity().getType());
		DiffCatResult result = makeResult(leftParent, rightParent, leftStatement, rightStatement, op);
		return result;
	}
	
	private ASTNode findRegularStatement(SourceCodeEntity changed, MethodDeclaration parent, ASTNode searchParent) {
		
		ANodeFinder rightFinder = mapStatementFinder(changed, searchParent);
		ASTNode rightStatement = discover(searchParent, rightFinder);
		if (rightStatement == null) {
			rightFinder = mapStatementFinder(changed, parent);
			rightStatement = discover (parent, rightFinder);
		}
		return rightStatement;
	}

	private ASTNode findRegularStatement(SourceCodeChange change, SourceCodeEntity changed, MethodDeclaration parent, ASTNode searchParent) {
		
		ANodeFinder rightFinder = mapStatementFinder(changed, change, searchParent);
		ASTNode rightStatement = discover(searchParent, rightFinder);
		if (rightStatement == null) {
			rightFinder = mapStatementFinder(changed, change, parent);
			rightStatement = discover (parent, rightFinder);
		}
		return rightStatement;
	}
	
	private ASTNode findParentStatement(SourceCodeEntity parentEntity, MethodDeclaration parent) {
		ASTNode searchParent = null;
		if (isAtStatementLevel(parentEntity)) {
			ANodeFinder rightFinder = mapStatementFinder(parentEntity, parent);
			searchParent = discover (parent, rightFinder);
		}
		else {
			searchParent = parent;
		}
		
		if (searchParent == null) {
			searchParent = parent;		// status quo
		}
		searchParent = mapToScopeParent(searchParent);
		return searchParent;
	}

	private boolean isAtStatementLevel(SourceCodeEntity parentEntity) {
		return EntityType.isAtStatementLevel(parentEntity.getType()) ||
		 		parentEntity.getType().equals(EntityType.THEN_STATEMENT);
	}

	private ASTNode mapToScopeParent (ASTNode node) {
		while (! isScopeParent(node)) {
			node = node.getParent();
		}
		
		return node;
	}
	
	private boolean isScopeParent (ASTNode node) {
		int nodeKind = node.getNodeType();
		switch (nodeKind) {
		case ASTNode.CATCH_CLAUSE:
		case ASTNode.ENHANCED_FOR_STATEMENT:
		case ASTNode.IF_STATEMENT:
		case ASTNode.METHOD_DECLARATION:
		case ASTNode.FOR_STATEMENT:
		case ASTNode.WHILE_STATEMENT:
		case ASTNode.DO_STATEMENT:
		case ASTNode.TRY_STATEMENT:
		case ASTNode.SWITCH_STATEMENT:
		case ASTNode.SYNCHRONIZED_STATEMENT:
		case ASTNode.TYPE_DECLARATION:
			return true;
		default:
			return false;
		}
	}
	
	private DiffCatResult mapAlternativePartInsertChange ( Insert change) {
		StructureEntityVersion root = change.getRootEntity();
	
		MethodDeclaration leftParent = getLeftMethodDeclaration(root);
		MethodDeclaration rightParent = getRightMethodDeclaration(root);
		
		ANodeFinder rightFinder = new ElseBlockFinder(change.getChangedEntity(), rightParent);
		ASTNode rightStatement = discover(rightParent, rightFinder);
		
		ChangeOperation op = ChangeOperation.ALTERNATIVE_PART_INSERT;
		return makeResult(leftParent, rightParent, null, rightStatement, op);
	}

	private DiffCatResult mapAlternativePartDeleteChange(  Delete change) {
		
		StructureEntityVersion root = change.getRootEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(root);
		MethodDeclaration rightParent = getRightMethodDeclaration(root);
		
		ANodeFinder leftFinder = new ElseBlockFinder(change.getChangedEntity(), leftParent);
		ASTNode leftStatement = discover(leftParent, leftFinder);
		
		ChangeOperation op = ChangeOperation.ALTERNATIVE_PART_DELETE;
		return makeResult(leftParent, rightParent, leftStatement, null, op);
	}

	private DiffCatResult mapAdditnlFuncltyChange ( Insert change )
	{
		StructureEntityVersion root = change.getRootEntity();
		
		ASTNode leftParent = getLeftTypeDeclaration(root);
		ASTNode rightParent = getRightTypeDeclaration(root);
		
		MethodDeclarationFinder rightMethodFinder = new MethodDeclarationFinder(change.getChangedEntity());
		MethodDeclaration rightNode = (MethodDeclaration) discover(rightParent, rightMethodFinder);
		
		// This saves us on jEdit #1429, but we'll probably die in many other cases if CD makes 
		// the same mistake as it does there
		if (rightNode == null)
		{
			HeuristicMethodInsertFinder f = new HeuristicMethodInsertFinder(change.getChangedEntity());
			rightParent.accept(f);
			List<MethodDeclaration> secondaries = f.getSecondaries();
			if (secondaries.size() == 1) {
				rightNode = secondaries.get(0);
			}
		}
		
		DiffCatResult result = makeResult(leftParent, rightParent, null, rightNode, ChangeOperation.FUNCTIONALITY_INSERT);
		return result;
	}
	
	private DiffCatResult mapRemovedFunctlyChange ( Delete change)
	{
		StructureEntityVersion root = change.getRootEntity();
		
		ASTNode leftParent = getLeftTypeDeclaration(root);
		ASTNode rightParent = getRightTypeDeclaration(root);
		
		MethodDeclaration leftNode = getLeftMethodDeclaration(change.getChangedEntity());
		
		DiffCatResult result = makeResult(leftParent, rightParent, leftNode, null, ChangeOperation.FUNCTIONALITY_DELETE);
		
		return result;
	}
	
	private DiffCatResult mapReturnTypeChange ( Update change ) {
		
		MethodDeclaration leftNode = getLeftMethodDeclaration(change.getRootEntity());
		MethodDeclaration rightNode = getRightMethodDeclaration(change.getRootEntity());
		
		// XXX ChangeDistiller fucks up on jEdit#2733 ... this hack fixes it
		if (leftNode.isConstructor() || rightNode.isConstructor()) {
			return null;
		}
		
		Type leftReturnType = leftNode.getReturnType2();
		Type rightReturnType = rightNode.getReturnType2();
		
		return makeResult(leftNode, rightNode, leftReturnType, rightReturnType, ChangeOperation.RETURN_TYPE_UPDATE);
	}
	
	private DiffCatResult mapMethodRenameChange(  Update change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(rootEntity);
		SimpleName left = leftParent.getName();
		
		MethodDeclaration rightParent = getRightMethodDeclaration(rootEntity);
		SimpleName right = rightParent.getName();
		
		return makeResult(leftParent, rightParent, left, right, ChangeOperation.METHOD_NAME_UPDATE);
	}

	private DiffCatResult mapParamDeleteChange ( Delete change)
	{
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(rootEntity);
		ASTNode left = ASTUtil.getParameter(CDUtil.getParamName(change.getChangedEntity()), leftParent);
		
		MethodDeclaration rightParent = getRightMethodDeclaration(rootEntity);
		
		return makeResult(leftParent, rightParent, left, null, ChangeOperation.PARAMETER_DELETE);
	}
	
	private DiffCatResult mapParamInsertChange( Insert change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(rootEntity);
		
		MethodDeclaration rightParent = getRightMethodDeclaration(rootEntity);
		ASTNode rightVar = ASTUtil.getParameter(CDUtil.getParamName(change.getChangedEntity()), rightParent);
		
		return makeResult(leftParent, rightParent, null, rightVar, ChangeOperation.PARAMETER_INSERT);
	}
	
	private DiffCatResult mapParamOrderingChange(  Move change) {
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(change.getRootEntity());
		MethodDeclaration rightParent = getRightMethodDeclaration(change.getRootEntity());
		
		ASTNode left = ASTUtil.getParameter(CDUtil.getParamName(change.getChangedEntity()), leftParent);
		ASTNode rightVar = ASTUtil.getParameter(CDUtil.getParamName(change.getNewEntity()), rightParent);
		
		ChangeOperation op = ChangeOperation.PARAMETER_ORDERING_UPDATE;
		
		return makeResult (leftParent, rightParent, left, rightVar, op);
	}
	
	private DiffCatResult mapParamRenameChange( Update change) {
		
		MethodDeclaration leftParent = getLeftMethodDeclaration(change.getRootEntity());
		MethodDeclaration rightParent = getRightMethodDeclaration(change.getRootEntity());
		
		ASTNode left = ASTUtil.getParameter(CDUtil.getParamName(change.getChangedEntity()), leftParent).getName();
		ASTNode rightVar = ASTUtil.getParameter(CDUtil.getParamName(change.getNewEntity()), rightParent).getName();
		
		ChangeOperation op = ChangeOperation.PARAMETER_NAME_UPDATE;
		return makeResult (leftParent, rightParent, left, rightVar, op);
	}
	
	private DiffCatResult mapParamTypeChange ( Update change)
	{
		MethodDeclaration leftParent = getLeftMethodDeclaration(change.getRootEntity());
		MethodDeclaration rightParent = getRightMethodDeclaration(change.getRootEntity());
		
		ASTNode left = ASTUtil.getParameter(CDUtil.getParamName(change.getChangedEntity()), leftParent).getType();
		ASTNode rightVar = ASTUtil.getParameter(CDUtil.getParamName(change.getNewEntity()), rightParent).getType();
		
		ChangeOperation op = ChangeOperation.PARAMETER_TYPE_UPDATE;
		return makeResult (leftParent, rightParent, left, rightVar, op);
	}
	
	private DiffCatResult mapCommentMoveChange( Move change) {
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		BodyDeclaration leftParent = getLeftBodyDeclaration(rootEntity);
		BodyDeclaration rightParent = getRightBodyDeclaration(rootEntity);
		
		String uniqueName = change.getChangedEntity().getUniqueName();
		ChangeOperation op = ChangeOperation.COMMENT_MOVE;
		SourceRange leftRange = change.getChangedEntity().getSourceRange();
		SourceRange rightRange = change.getNewEntity().getSourceRange();
		return makeCommentResult(leftParent, rightParent, uniqueName, uniqueName, op, leftRange, rightRange);
	}
	
	private DiffCatResult mapCommentUpdateChange( Update change) {
		
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		ASTNode leftParent = getLeftBodyDeclaration(rootEntity);
		ASTNode rightParent = getRightBodyDeclaration(rootEntity);
		
		String uniqueNameLeft = change.getChangedEntity().getUniqueName();
		String uniqueNameRight = change.getNewEntity().getUniqueName();
		
		SourceRange leftRange = change.getChangedEntity().getSourceRange();
		SourceRange rightRange = change.getNewEntity().getSourceRange();
		
		ChangeOperation op = ChangeOperation.COMMENT_UPDATE;
		
		return makeCommentResult(leftParent, rightParent, uniqueNameLeft, uniqueNameRight, op, leftRange, rightRange);
	}
	
	private DiffCatResult mapCommentInsertChange( Insert change) {
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		ASTNode leftParent = getLeftBodyDeclaration( rootEntity);
		ASTNode rightParent = getRightBodyDeclaration( rootEntity);
		
		String uniqueName = change.getChangedEntity().getUniqueName();
		ChangeOperation op = ChangeOperation.COMMENT_INSERT;
		SourceRange rightRange = change.getChangedEntity().getSourceRange();
		return makeCommentResult(leftParent, rightParent, null, uniqueName, op, null, rightRange);
	}
	
	private DiffCatResult mapCommentDeleteChange( Delete change) {
		StructureEntityVersion rootEntity = change.getRootEntity();
		
		ASTNode leftParent = getLeftBodyDeclaration( rootEntity);
		ASTNode rightParent = getRightBodyDeclaration( rootEntity);
		
		String uniqueName = change.getChangedEntity().getUniqueName();
		ChangeOperation op = ChangeOperation.COMMENT_DELETE;
		SourceRange leftRange = change.getChangedEntity().getSourceRange();
		return makeCommentResult( leftParent, rightParent, uniqueName, null, op, leftRange, null);
	}
	
	private DiffCatResult mapJavadocUpdateChange (Update change)
	{
		StructureEntityVersion root = change.getRootEntity();
		
		BodyDeclaration leftParent = getLeftBodyDeclaration(root);
		BodyDeclaration rightParent = getRightBodyDeclaration(root);
		
		ASTNode leftDoc = leftParent != null? leftParent.getJavadoc() : null;
		ASTNode rightDoc = rightParent != null? rightParent.getJavadoc() : null;
		
		return makeResult(leftParent, rightParent, leftDoc, rightDoc, ChangeOperation.DOC_UPDATE);
	}
	
	private BodyDeclaration getLeftBodyDeclaration (StructureEntityVersion root) {
		if (root.getType() == EntityType.CLASS || root.getType() == EntityType.TYPE_DECLARATION || root.getType() == EntityType.ENUM_DECLARATION) {
			return getLeftTypeDeclaration (root);
		}
		else if (root.getType() == EntityType.METHOD || root.getType() == EntityType.METHOD_DECLARATION) {
			return getLeftMethodDeclaration(root);
		}
		else {
			return getLeftFieldDeclaration(root);
		}
	}
	
	private BodyDeclaration getRightBodyDeclaration (StructureEntityVersion root) {
		if (root.getType() == EntityType.CLASS || root.getType() == EntityType.TYPE_DECLARATION || root.getType() == EntityType.ENUM_DECLARATION) {
			return getRightTypeDeclaration(root);
		}
		else if (root.getType() == EntityType.METHOD || root.getType() == EntityType.METHOD_DECLARATION) {
			return getRightMethodDeclaration(root);
		}
		else {
			return getRightFieldDeclaration(root);
		}
	}
	
	private AbstractTypeDeclaration getLeftTypeDeclaration (StructureEntityVersion root) {
		if (classRenamesBackward.containsKey(root.toString())) {
			return classRenamesBackward.get(root.toString());
		}
		
		TypeDeclarationFinder leftFinder = new TypeDeclarationFinder(root);
		return (AbstractTypeDeclaration) discover (leftType, leftFinder);
	}
	
	private AbstractTypeDeclaration getRightTypeDeclaration (StructureEntityVersion root) {
		TypeDeclarationFinder rightFinder = new TypeDeclarationFinder(root);
		return (AbstractTypeDeclaration) discover (rightType, rightFinder);
	}
	
	private FieldDeclaration getLeftFieldDeclaration (StructureEntityVersion root) {
		if (fieldSignatureUpdates.containsKey(root.toString())) {
			return fieldSignatureUpdates.get(root.toString());
		}
		
		FieldDeclarationFinder leftFinder = new FieldDeclarationFinder(root);
		return (FieldDeclaration) discover (leftType, leftFinder);
	}
	
	private FieldDeclaration getRightFieldDeclaration (StructureEntityVersion root) {
		FieldDeclarationFinder rightFinder = new FieldDeclarationFinder(root);
		return (FieldDeclaration) discover (rightType, rightFinder);
	}
	
	private MethodDeclaration getLeftMethodDeclaration (StructureEntityVersion root) {
		if (methodSignatureUpdates.containsKey(root.toString())) {
			return methodSignatureUpdates.get(root.toString());
		}
		
		MethodDeclarationFinder leftFinder = new MethodDeclarationFinder(root);
		MethodDeclaration declaration = (MethodDeclaration) discover(leftType, leftFinder);
		
		if (declaration != null) {
			return declaration;
		}
		
		// if null, might be []'ing issue with ChangeDistiller
		SoftMethodDeclarationFinder finder = new SoftMethodDeclarationFinder(root);
		declaration = (MethodDeclaration) discover(leftType, finder);
		
		return declaration;
	}
	
	private MethodDeclaration getLeftMethodDeclaration (SourceCodeEntity root) {
		if (methodSignatureUpdates.containsKey(root.toString())) {
			return methodSignatureUpdates.get(root.toString());
		}
		
		MethodDeclarationFinder leftFinder = new MethodDeclarationFinder(root);
		MethodDeclaration declaration = (MethodDeclaration) discover(leftType, leftFinder);
		
		if (declaration != null) {
			return declaration;
		}
		
		// if null, might be []'ing issue with ChangeDistiller
		SoftMethodDeclarationFinder finder = new SoftMethodDeclarationFinder(root);
		declaration = (MethodDeclaration) discover(leftType, finder);
		
		return declaration;
	}
	
	private MethodDeclaration getRightMethodDeclaration ( StructureEntityVersion root) {
		
		MethodDeclarationFinder finder = new MethodDeclarationFinder(root);
		MethodDeclaration declaration = (MethodDeclaration) discover(rightType, finder);
		
		if (declaration != null) {
			return declaration;
		}
		
		SoftMethodDeclarationFinder rightFinder = new SoftMethodDeclarationFinder(root);
		declaration = (MethodDeclaration) discover(rightType, rightFinder);
		
		return declaration;
	}
	
	private ASTNode discover (ASTNode toVisit, ANodeFinder finder) {
		toVisit.accept(finder);
		return finder.getAssociatedNode();
	}
	
	private DiffCatResult makeCommentResult( ASTNode leftParent, ASTNode rightParent, String leftComment,
												String rightComment, ChangeOperation op, SourceRange leftRange, SourceRange rightRange)
	{
		DiffCatResult result = makeResult(leftParent, rightParent, op);
		if (leftComment != null) {
			result.setHandleFrom(leftComment.toString());
			result.setPosFrom(leftRange.getOffset());
			result.setLenFrom(leftRange.getLength());
		}
		if (rightComment != null) {
			result.setHandleTo(rightComment.toString());
			result.setPosTo(rightRange.getOffset());
			result.setLenTo(rightRange.getLength());
		}
		return result;
	}
	
	private DiffCatResult makeResult( ASTNode leftParent, ASTNode rightParent, ChangeOperation op) {
		return makeResult (leftParent, rightParent, null, null, op);
	}
	
	private DiffCatResult makeResult( ASTNode leftParent, ASTNode rightParent, ASTNode leftNode,
										ASTNode rightNode, ChangeOperation op) 
	{
		DiffCatResult result = new DiffCatResult();
		result.setChangeType(op.name());
		result.setLeftParent(leftParent);
		result.setRightParent(rightParent);
		result.setNodeFrom(leftNode);
		result.setNodeTo(rightNode);
		
		return result;
	}
	
	private ANodeFinder mapStatementFinder (SourceCodeEntity changed, ASTNode parent) {
		if (changed.getType() == EntityType.CATCH_CLAUSE) {
			return new CatchClauseFinder(changed, parent);
		}
		if (changed.getType() == EntityType.TRY_STATEMENT) {
			return new TryStatementFinder(changed, parent);
		}
		if (changed.getType() == EntityType.RETURN_STATEMENT) {
			return new ReturnStatementFinder(changed, parent);
		}
		if (changed.getType() == EntityType.FOR_STATEMENT) {
			return new ForStatementFinder(changed, parent);
		}
		if (changed.getType() == EntityType.BREAK_STATEMENT) {
			return new BreakStatementFinder(changed, parent);
		}
		if (changed.getType() == EntityType.CONTINUE_STATEMENT) {
			return new ContinueStatementFinder(changed, parent);
		}
		if (changed.getType() == EntityType.LABELED_STATEMENT) {
			return new LabelledStatementFinder(changed, parent);
		}
		
		return new StatementFinder(changed, parent);
	}
	
	private ANodeFinder mapStatementFinder (SourceCodeEntity changed, SourceCodeChange change, ASTNode parent) {
		if (change.getChangeType().equals(ChangeType.CONDITION_EXPRESSION_CHANGE)) {
			return new ConditionFinder(changed, parent);
		}
		
		return new StatementFinder(changed, parent);
	}
}
