/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.cd;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.evolizer.changedistiller.treedifferencing.matching.measure.LevenshteinCalculator;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;

public class FieldOpDetector {

	private static final int MAX_LENGTH = 1000;

	private static FieldOpDetector singleton = new FieldOpDetector();
	
	public static FieldOpDetector getDetector () {
		return singleton;
	}
	
	private final double MIN_SIM_THRESH = 0.65;
	
	private final LevenshteinCalculator calculator = new LevenshteinCalculator();
	
	private AbstractTypeDeclaration left;
	private AbstractTypeDeclaration right;
	
	private FieldOpDetector () {
		
	}
	
	public Set<DiffCatResult> getAttributeOps (AbstractTypeDeclaration left, AbstractTypeDeclaration right) {
		
		this.left = left;
		this.right = right;
		
		// results
		Set<DiffCatResult> results = new HashSet<DiffCatResult>();
		
		List<VariableDeclarationFragment> leftFrags = collectFieldDeclarations(left);
		List<VariableDeclarationFragment> rightFrags = collectFieldDeclarations(right);
		
		// maps left field decls to the best right field decl
		HashMap<VariableDeclarationFragment, BestResult> leftResults 
			= new HashMap<VariableDeclarationFragment, FieldOpDetector.BestResult>();
		
		// maps right field decls to the best left field decl
		HashMap<VariableDeclarationFragment, BestResult> rightResults
			= new HashMap<VariableDeclarationFragment, FieldOpDetector.BestResult>();
		
		// map left field decls
		for (VariableDeclarationFragment leftFrag : leftFrags) {
			
			double bestLevenshtein = 0.0;
			VariableDeclarationFragment bestRightFrag = null;
			
			for (VariableDeclarationFragment rightFrag : rightFrags) {
				double tempLevenshtein = getLevenshtein(leftFrag, rightFrag);
				if (tempLevenshtein > bestLevenshtein) {
					bestLevenshtein = tempLevenshtein;
					bestRightFrag = rightFrag;
				}
			}
			
			leftResults.put(leftFrag, new BestResult(bestLevenshtein, bestRightFrag));
		}
		
		// map right field decls
		for (VariableDeclarationFragment rightFrag : rightFrags) {
			
			double bestLevenshtein = 0.0;
			VariableDeclarationFragment bestLeftFrag = null;
			
			for (VariableDeclarationFragment leftFrag : leftFrags) {
				double tempLevenshtein = getLevenshtein(leftFrag, rightFrag);
				if (tempLevenshtein > bestLevenshtein) {
					bestLevenshtein = tempLevenshtein;
					bestLeftFrag = leftFrag;
				}
			}
			
			rightResults.put(rightFrag, new BestResult(bestLevenshtein, bestLeftFrag));
		}
		
		// link up lefts with rights
		for (VariableDeclarationFragment leftFrag : leftFrags) 
		{
			BestResult leftResult = leftResults.get(leftFrag);
			
			// if top result < threshold, then say the field was deleted
			if (leftResult.levenshtein < MIN_SIM_THRESH) {
				results.add(makeDelete(leftFrag));
				leftResults.remove(leftFrag);
				continue;
			}
			
			// otherwise, check if the right has the same left 
			BestResult rightResult = rightResults.get(leftResult.frag);
			
			// already taken?
			if (rightResult == null) {
				results.add(makeDelete(leftFrag));
				leftResults.remove(leftFrag);
				continue;
			}
			
			if (rightResult.frag != leftFrag) {
				results.add(makeDelete(leftFrag));
				leftResults.remove(leftFrag);
				continue;
			}
			
			// if so, then we create attribute update results
			results.addAll(makeUpdates(leftFrag, leftResult.frag));
			
			// delete left and right
			leftResults.remove(leftFrag);
			rightResults.remove(leftResult.frag);
		}
		
		// link up remaining rights with lefts
		for (VariableDeclarationFragment rightFrag : rightFrags) 
		{
			BestResult rightResult = rightResults.get(rightFrag);
			
			// already dealt with this one
			if (rightResult == null) {
				continue;
			}
			
			// if top result < threshold, then say the field was deleted
			if (rightResult.levenshtein < MIN_SIM_THRESH) {
				results.add(makeInsert(rightFrag));
				rightResults.remove(rightFrag);
				continue;
			}
			
			// otherwise, check if the right has the same left 
			BestResult leftResult = leftResults.get(rightResult.frag);
			
			if (leftResult == null || leftResult.frag != rightFrag) {
				results.add(makeInsert(rightFrag));
				rightResults.remove(rightFrag);
				continue;
			}
			
			// if so, then we create attribute update results
			results.addAll(makeUpdates(rightResult.frag, rightFrag));
			
			// delete left and right
			leftResults.remove(rightResult.frag);
			rightResults.remove(rightFrag);
		}
		
		// do inner types with same names
		Set<ASTNode> leftMemberTypes = ASTUtil.getMemberTypes(left);
		Set<ASTNode> rightMemberTypes = ASTUtil.getMemberTypes(right);
		
		for (ASTNode leftInner : leftMemberTypes) {
			AbstractTypeDeclaration leftMemberType = (AbstractTypeDeclaration) leftInner;
			String leftName = leftMemberType.getName().getIdentifier();
			for (ASTNode rightInner : rightMemberTypes) {
				AbstractTypeDeclaration rightMemberType = (AbstractTypeDeclaration) rightInner;
				String rightName = rightMemberType.getName().getIdentifier();
				if (leftName.equals(rightName)) {
					FieldOpDetector temp = new FieldOpDetector();
					results.addAll(temp.getAttributeOps(leftMemberType, rightMemberType));
				}
			}
		}
		
		this.left = null;
		this.right = null;
		
		return results;
	}
	
	private DiffCatResult makeDelete (VariableDeclarationFragment leftFrag)
	{
		DiffCatResult result = new DiffCatResult();
		result.setChangeType(ChangeOperation.ATTRIBUTE_DELETE.name());
		result.setNodeFrom(leftFrag);
		result.setLeftParent(left);
		result.setRightParent(right);
		
		return result;
	}
	
	private DiffCatResult makeInsert (VariableDeclarationFragment rightFrag)
	{
		DiffCatResult result = new DiffCatResult();
		result.setChangeType(ChangeOperation.ATTRIBUTE_INSERT.name());
		result.setNodeTo(rightFrag);
		result.setLeftParent(left);
		result.setRightParent(right);
		
		return result;
	}
	
	private Set<DiffCatResult> makeUpdates (VariableDeclarationFragment left, VariableDeclarationFragment right)
	{
		Set<DiffCatResult> updates = new HashSet<DiffCatResult>();
		
		SimpleName leftName = left.getName();
		SimpleName rightName = right.getName();
		
		String leftStringName = leftName.getIdentifier();
		String rightStringName = rightName.getIdentifier();
		
		if (! leftStringName.equals(rightStringName)) {
			updates.add(makeRename(left, right));
		}
		
		FieldDeclaration leftParent = (FieldDeclaration) left.getParent();
		Type leftType = leftParent.getType();
		
		FieldDeclaration rightParent = (FieldDeclaration) right.getParent();
		Type rightType = rightParent.getType();
		
		ITypeBinding leftBinding = leftType.resolveBinding();
		ITypeBinding rightBinding = rightType.resolveBinding();
		
		if (! BindingUtil.areEqual(leftBinding, rightBinding)) {
			if (leftName.resolveTypeBinding() != null && rightName.resolveTypeBinding() != null) {
				ITypeBinding type1 = ((IVariableBinding) leftName.resolveBinding()).getType();
				ITypeBinding type2 = ((IVariableBinding) rightName.resolveBinding()).getType();
				
				// special case: int x[] --> int[] x is a trivial type update 
				if (BindingUtil.areEqual(type1, type2)) {
					DiffCatResult makeTypeUpdate = makeTypeUpdate(leftType, left, rightType, right);
					makeTypeUpdate.setChangeType(ChangeOperation.ATTRIBUTE_TYPE_TRIVIAL_UPDATE.name());
					updates.add(makeTypeUpdate);
				}
			}
		}
		
		Expression leftExpr = left.getInitializer();
		Expression rightExpr = right.getInitializer();
		
		if (leftExpr == null && rightExpr != null) {
			updates.add(makeExpressionInsert(left, right));
		}
		else if (leftExpr != null && rightExpr == null) {
			updates.add(makeExpressionDelete(left, right));
		}
		else if (leftExpr != null && rightExpr != null){
			String leftExprString = leftExpr.toString();
			String rightExprString = rightExpr.toString();
			
			if (! leftExprString.equals(rightExprString)) {
				updates.add(makeExpressionUpdate(left, right));
			}
		}
		
		int leftModifiers = leftParent.getModifiers();
		int rightModifiers = rightParent.getModifiers();
		
		updates.addAll(getModifierResults(leftModifiers, left, rightModifiers, right));
		
		return updates;
	}

	private Set<DiffCatResult> getModifierResults (int leftMod, VariableDeclarationFragment left, int rightMod, VariableDeclarationFragment right) {
		
		Set<DiffCatResult> modifierResults = new HashSet<DiffCatResult>();
		
		// modifiability checking
		if (Modifier.isFinal(leftMod) && ! Modifier.isFinal(rightMod)) {
			modifierResults.add(makeGenericChange(left, right, ChangeOperation.ATTRIBUTE_MODIFIABILITY_INSERT.name()));
		}
		else if (! Modifier.isFinal(leftMod) && Modifier.isFinal(rightMod)) {
			modifierResults.add(makeGenericChange(left, right, ChangeOperation.ATTRIBUTE_MODIFIABILITY_DELETE.name()));
		}
		
		// accessibility checking
		boolean leftModBool = isDefault (leftMod);
		boolean rightModBool = isDefault (rightMod);
		
		if (leftModBool && ! rightModBool) {
			modifierResults.add(makeGenericChange(left, right, ChangeOperation.ATTRIBUTE_ACCESSIBILITY_DELETE.name()));
		}
		else if (! leftModBool && rightModBool) {
			modifierResults.add(makeGenericChange(left, right, ChangeOperation.ATTRIBUTE_ACCESSIBILITY_INSERT.name()));
		}
		else {
			if (Modifier.isPrivate(leftMod) != Modifier.isPrivate(rightMod)) {
				modifierResults.add(makeGenericChange(left, right, ChangeOperation.ATTRIBUTE_ACCESSIBILITY_UPDATE.name()));
			}
			else if (Modifier.isProtected(leftMod) != Modifier.isProtected(rightMod)) {
				modifierResults.add(makeGenericChange(left, right, ChangeOperation.ATTRIBUTE_ACCESSIBILITY_UPDATE.name()));
			}
			else if (Modifier.isPublic(leftMod) != Modifier.isPublic(rightMod)) {
				modifierResults.add(makeGenericChange(left, right, ChangeOperation.ATTRIBUTE_ACCESSIBILITY_UPDATE.name()));
			}
		}
		
		return modifierResults;
	}
	
	private DiffCatResult makeGenericChange(VariableDeclarationFragment left2, VariableDeclarationFragment right2, String name) {
		DiffCatResult result = new DiffCatResult();
		result.setChangeType(name);
		result.setNodeFrom(left2);
		result.setNodeTo(right2);
		result.setLeftParent(left2.getParent());
		result.setRightParent(right2.getParent());
		return result;
	}

	private boolean isDefault (int mod) {
		 return ! Modifier.isPrivate(mod) && ! Modifier.isProtected(mod) && ! Modifier.isPublic(mod);
	}
	
	private DiffCatResult makeExpressionUpdate( VariableDeclarationFragment left2, VariableDeclarationFragment right2) {
		DiffCatResult result = new DiffCatResult();
		result.setChangeType(ChangeOperation.ATTRIBUTE_INIT_UPDATE.name());
		result.setNodeFrom(left2.getInitializer());
		result.setNodeTo(right2.getInitializer());
		result.setLeftParent(left2);
		result.setRightParent(right2);
		return result;
	}

	private DiffCatResult makeExpressionDelete( VariableDeclarationFragment left2, VariableDeclarationFragment right2) {
		DiffCatResult result = new DiffCatResult();
		result.setChangeType(ChangeOperation.ATTRIBUTE_INIT_DELETE.name());
		result.setNodeFrom(left2.getInitializer());
		result.setLeftParent(left2);
		result.setRightParent(right2);
		return result;
	}

	private DiffCatResult makeExpressionInsert( VariableDeclarationFragment left2, VariableDeclarationFragment right2) {
		DiffCatResult result = new DiffCatResult();
		result.setChangeType(ChangeOperation.ATTRIBUTE_INIT_INSERT.name());
		result.setNodeTo(right2.getInitializer());
		result.setLeftParent(left2);
		result.setRightParent(right2);
		return result;
	}

	private DiffCatResult makeTypeUpdate(Type leftType, VariableDeclarationFragment left2, Type rightType, VariableDeclarationFragment right2) {
		 
		DiffCatResult update = new DiffCatResult();
		update.setChangeType(ChangeOperation.ATTRIBUTE_TYPE_UPDATE.name());
		update.setNodeFrom(leftType);
		update.setNodeTo(rightType);
		update.setLeftParent(left2);
		update.setRightParent(right2);
		
		return update;
	}

	private DiffCatResult makeRename(VariableDeclarationFragment left2, VariableDeclarationFragment right2) {
		
		DiffCatResult rename = new DiffCatResult();
		rename.setChangeType(ChangeOperation.ATTRIBUTE_NAME_UPDATE.name());
		rename.setNodeFrom(left2.getName());
		rename.setNodeTo(right2.getName());
		rename.setLeftParent(left2);
		rename.setRightParent(right2);
		return rename;
	}

	private double getLevenshtein (VariableDeclarationFragment left, VariableDeclarationFragment right) {
		
		if (BindingUtil.areEqual(left.resolveBinding(), right.resolveBinding())) {
			return 1.0;	// same bindings should always be equal! [April 6, 2011]
		}
		
		String leftID = getString (left);
		String rightID = getString (right);
		
		if (leftID.length() < MAX_LENGTH && rightID.length() < MAX_LENGTH) {
			return calculator.calculateSimilarity(leftID, rightID);
		}
		else {
			return leftID.equals(rightID) ? 1.0 : 0.0;
		}
	}
	
	private String getString (VariableDeclarationFragment frag) {
		StringBuffer buf = new StringBuffer();
		buf.append(((FieldDeclaration) frag.getParent()).getType().toString());
		buf.append(" ");
		buf.append(frag.getName().getIdentifier());
		buf.append(" ");
		buf.append(frag.getInitializer() != null ? frag.getInitializer().toString() : "");
		return buf.toString();
	}
	
	@SuppressWarnings("rawtypes")
	private List<VariableDeclarationFragment> collectFieldDeclarations (AbstractTypeDeclaration type)
	{
		List<VariableDeclarationFragment> fields = new LinkedList<VariableDeclarationFragment>();
		
		List bodyDecls = type.bodyDeclarations();
		for (Object bodyDecl : bodyDecls) {
			
			if (bodyDecl instanceof FieldDeclaration) {
				FieldDeclaration fieldDecl = (FieldDeclaration) bodyDecl;
				
				List varFrags = fieldDecl.fragments();
				
				for (Object varFrag : varFrags) {
					
					fields.add((VariableDeclarationFragment) varFrag);
				}
			}
			
		}
		
		return fields;
	}
	
	private class BestResult {
		double levenshtein;
		VariableDeclarationFragment frag;
		
		BestResult(double levensthein, VariableDeclarationFragment frag) {
			this.levenshtein = levensthein;
			this.frag = frag;
		}
		
		public String toString () {
			return frag + " " + levenshtein;
		}
	}
}
