/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.cd;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Name;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.model.DiffCodes;
import ca.mcgill.cs.swevo.diffcat.util.DiffCatParentMap;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.LRTupleSet;
import ca.mcgill.cs.swevo.util.SetUtil;

public class ThrownExceptionsDetector {

	private static ThrownExceptionsDetector singleton = new ThrownExceptionsDetector();
	
	public static ThrownExceptionsDetector getDetector () {
		return singleton;
	}
	
	private ThrownExceptionsDetector () {
		
	}
	
	@SuppressWarnings("rawtypes")
	public Set<DiffCatResult> getThrownExceptionUpdates (AbstractTypeDeclaration left, AbstractTypeDeclaration right, Set<DiffCatResult> knowns)
	{
		Set<DiffCatResult> results = new HashSet<DiffCatResult>();
		
		LRTupleSet<ASTNode> methodSigNodes = null;
		
		// use the knowns to get thrown exception updates for refactored signatures
		Set<DiffCatResult> methodSigUpdates = SetUtil.filterAndCopy(knowns, DiffCodes.METHOD_SIGNATURE_DIFF);
		DiffCatParentMap map = new DiffCatParentMap(methodSigUpdates);
		methodSigNodes = map.keySet();
		
		// add all other matching signatures
		List<MethodDeclaration> leftMethods = ASTUtil.getMethods(left);
		List<MethodDeclaration> rightMethods = ASTUtil.getMethods(right);
		
		for (MethodDeclaration leftMethod : leftMethods) {
			IMethodBinding leftBinding = null; 
			try {
				leftBinding = leftMethod.resolveBinding();
			} catch (NullPointerException e) {
				e.printStackTrace();
				continue;
			}
			if (leftBinding == null) {
				continue;
			}
			for (MethodDeclaration rightMethod : rightMethods) {
				IMethodBinding rightBinding = null;
				try {
					rightBinding = rightMethod.resolveBinding();
				} catch (NullPointerException e) {
					e.printStackTrace();
					continue;
				}
				if (rightBinding == null) {
					continue;
				}
				if (leftBinding.isConstructor() && rightBinding.isConstructor()) {
					if (BindingUtil.areEqual(leftBinding.getParameterTypes(), rightBinding.getParameterTypes())) {
						methodSigNodes.add(new LRTuple<ASTNode>(leftMethod, rightMethod));
						continue;
					}
				}
				if (BindingUtil.areWeaklyOverriding(leftBinding, rightBinding)) {
					methodSigNodes.add(new LRTuple<ASTNode>(leftMethod, rightMethod));
				}
			}
		}
		
		// get thrown exception diffs
		for (LRTuple<ASTNode> methodPair : methodSigNodes)
		{
			MethodDeclaration leftPair = (MethodDeclaration) methodPair.getLeft();
			MethodDeclaration rightPair = (MethodDeclaration) methodPair.getRight();
			
			if (leftPair == null || rightPair == null) {
				continue; // TODO Ant bug, 4448
			}
			
			List leftExceptions = leftPair.thrownExceptions();
			List rightExceptions = rightPair.thrownExceptions();
			
			// get deleted exceptions (text-based only)
			for (Object o1 : leftExceptions) {
				boolean isTextMatched = false;
				String leftException = o1.toString();
				for (Object o2 : rightExceptions) {
					String rightException = o2.toString();
					if (leftException.equals(rightException)) {
						isTextMatched = true;
					}
				}
				
				if (! isTextMatched) {
					results.add(makeThrownExceptionDelete((Name) o1, leftPair, rightPair));
				}
			}
			
			// get inserted exceptions (text-based only)
			for (Object o1 : rightExceptions) {
				boolean isTextMatched = false;
				String rightException = o1.toString();
				for (Object o2 : leftExceptions) {
					String leftException = o2.toString();
					if (rightException.equals(leftException)) {
						isTextMatched = true;
					}
				}
				
				if (! isTextMatched) {
					results.add(makeThrownExceptionInsert((Name) o1, leftPair, rightPair));
				}
			}
		}
		
		// do inner types with same names
		Set<ASTNode> leftMemberTypes = ASTUtil.getMemberTypes(left);
		Set<ASTNode> rightMemberTypes = ASTUtil.getMemberTypes(right);
		
		for (ASTNode leftInner : leftMemberTypes) {
			AbstractTypeDeclaration leftMemberType = (AbstractTypeDeclaration) leftInner;
			String leftName = leftMemberType.getName().getIdentifier();
			for (ASTNode rightInner : rightMemberTypes) {
				AbstractTypeDeclaration rightMemberType = (AbstractTypeDeclaration) rightInner;
				String rightName = rightMemberType.getName().getIdentifier();
				if (leftName.equals(rightName)) {
					results.addAll(getThrownExceptionUpdates(leftMemberType, rightMemberType, knowns));
				}
			}
		}
		
		return results;
	}
	
	private DiffCatResult makeThrownExceptionDelete (ASTNode delException, MethodDeclaration leftDecl, MethodDeclaration rightDecl)
	{
		DiffCatResult result = new DiffCatResult();
		result.setChangeType(ChangeOperation.THROWN_EXCEPTION_DELETE.name());
		result.setNodeFrom(delException);
		result.setLeftParent(leftDecl);
		result.setRightParent(rightDecl);
		return result;
	}
	
	private DiffCatResult makeThrownExceptionInsert (ASTNode insException, MethodDeclaration leftDecl, MethodDeclaration rightDeclaration)
	{
		DiffCatResult result = new DiffCatResult();
		result.setChangeType(ChangeOperation.THROWN_EXCEPTION_INSERT.name());
		result.setNodeTo(insException);
		result.setLeftParent(leftDecl);
		result.setRightParent(rightDeclaration);
		return result;
	}
}
