/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.cd.finder;

import java.util.Arrays;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclaration;
import org.evolizer.changedistiller.model.classifiers.ChangeType;
import org.evolizer.changedistiller.model.classifiers.EntityType;
import org.evolizer.changedistiller.model.entities.SourceCodeEntity;
import org.evolizer.changedistiller.model.entities.StructureEntityVersion;

import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;


public class CDUtil {

	private static final String COMMA = ",";
	private static final String LEFT_BRACKET = "(";
	private static final String FIELD_TYPE_SPLIT = ":";
	private static final String NON_REGEX_DOT = "\\.";
	private static final String VAR_ARGS_STRING = NON_REGEX_DOT + NON_REGEX_DOT + NON_REGEX_DOT;
	private static final String ARRAY_BRACKETS = "[]";
	
	public static String getParamName (SourceCodeEntity paramEntity)
	{
		String name = paramEntity.getUniqueName();
		String[] nameAndType = name.split(FIELD_TYPE_SPLIT);
		return nameAndType[0].trim();
	}
	
	public static String[] getParamType (SourceCodeEntity paramEntity)
	{
		String name = paramEntity.getUniqueName();
		String[] nameAndType = name.split(FIELD_TYPE_SPLIT);
		if (nameAndType.length <= 1) {
			return BindingUtil.UNKNOWN_TYPE_BINDING;
		}
		return nameAndType[1].trim().split(NON_REGEX_DOT);
	}
	
	public static String getFieldName (StructureEntityVersion fieldDecl) {
		return getFieldName(fieldDecl.getUniqueName().trim());
	}
	
	public static String getFieldName (SourceCodeEntity fieldDecl) {
		return getFieldName (fieldDecl.getUniqueName().trim());
	}
	
	public static String[] getFieldTypeFQN (SourceCodeEntity fieldDecl) {
		return getFieldTypeFQN (fieldDecl.getUniqueName().trim());
	}
	
	public static String getMethodName (SourceCodeEntity methodDecl) {
		return getMethodName( methodDecl.getUniqueName().trim() );
	}
	
	public static String getMethodName (StructureEntityVersion methodDecl) {
		return getMethodName( methodDecl.getUniqueName() );
	}
	
	public static String[] getFieldTypeFQN (String field)
	{
		int colonIndex = field.indexOf(FIELD_TYPE_SPLIT);
		
		if (colonIndex != -1) {
			field = field.substring(colonIndex+1).trim();
		}
		
		return field.split(NON_REGEX_DOT);
	}
	
	public static String getFieldName (String field)
	{
		int colonIndex = field.indexOf(FIELD_TYPE_SPLIT);
		
		if (colonIndex != -1) {
			field = field.substring(0, colonIndex);
		}
		
		String[] parts = field.split(NON_REGEX_DOT);
		
		if (parts.length == 0) {
			return "";
		}
		
		return parts[parts.length - 1].trim();
	}
	
	public static String getMethodName (String method)
	{
		method = BindingUtil.removeGenerics(method);
		method = BindingUtil.translateVarArgs(method);
		String[] parts = method.split(NON_REGEX_DOT);
		
		if (parts.length == 0) {
			return method;
		}
		
		String last = parts[parts.length - 1];
		int bracketIndex = last.indexOf(LEFT_BRACKET);
		
		if (bracketIndex == -1) {
			return last;
		}
		
		return last.substring(0, bracketIndex);
	}
	
	public static String[] getMethodDeclaringType (SourceCodeEntity methodDecl) {
		return getMethodDeclaringType( methodDecl.getUniqueName() );
	}
	
	public static String[] getMethodDeclaringType (StructureEntityVersion methodDecl) {
		return getMethodDeclaringType( methodDecl.getUniqueName() );
	}
	
	public static String[] getMethodDeclaringType (String method)
	{
		String[] split = method.split(NON_REGEX_DOT);
		if (split.length <= 1) {
			return split;
		}
		return Arrays.copyOfRange(split, 0, split.length - 1);
	}
	
	public static String[][] getMethodParameterTypes (SourceCodeEntity methodDecl) {
		return getMethodParameterTypes( methodDecl.getUniqueName() );
	}
	
	public static String[][] getMethodParameterTypes (StructureEntityVersion methodDecl) {
		return getMethodParameterTypes( methodDecl.getUniqueName() );
	}
	
	public static String[][] getMethodParameterTypesIgnoreBrackets (StructureEntityVersion methodDecl) {
		String[][] types = getMethodParameterTypes( methodDecl.getUniqueName() );
		for (int i = 0; i < types.length; i++) {
			types[i][types[i].length - 1] = BindingUtil.removeBrackets(types[i][types[i].length - 1]);
		}
		return types;
	}
	
	public static String[][] getMethodParameterTypesIgnoreBrackets (SourceCodeEntity methodDecl) {
		String[][] types = getMethodParameterTypes( methodDecl.getUniqueName() );
		for (int i = 0; i < types.length; i++) {
			types[i][types[i].length - 1] = BindingUtil.removeBrackets(types[i][types[i].length - 1]);
		}
		return types;
	}
	
	public static String[][] getMethodParameterTypes (String method)
	{
		String whole = method;
		int bracket = whole.indexOf(LEFT_BRACKET);
		if (bracket == -1) {
			return new String[0][];
		}
		String argString = whole.substring(bracket+1, whole.length() - 1);
		if (argString.isEmpty()) {
			return new String[0][];
		}
		argString = BindingUtil.removeGenerics (argString);
		argString = BindingUtil.translateVarArgs(argString);
		String[] args = argString.split(COMMA);
		String[][] result = new String[args.length][];
		for (int i = 0; i < args.length; i++) {
			String arg = args[i].trim();
			result[i] = arg.split(NON_REGEX_DOT);
			
		}
		return result;
	}
	
	public static String[] getTypeFQN (SourceCodeEntity typeDecl) {
		return typeDecl.getUniqueName().split(NON_REGEX_DOT);
	}
	
	public static String[] getTypeFQN (StructureEntityVersion typeDecl) {
		return typeDecl.getUniqueName().split(NON_REGEX_DOT);
	}
	
	public static String[] getTypeFQN (String s) {
		return s.split(NON_REGEX_DOT);
	}
}
