/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.reconciler;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.DiffCatParentMap;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsChangeOperationFilter;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.SetUtil;
import static ca.mcgill.cs.swevo.util.CollectionUtil.*;

/**
 * Detects those updates that are induced entirely by 
 * a rename operation.
 * 
 * @author dkawry
 *
 */

public class RenameReconciler extends AbstractReconciler {

	protected final IsChangeOperationFilter typeFilter = 
		new IsChangeOperationFilter
		( 
			ChangeOperation.VARIABLE_DECLARED_TYPE_UPDATE.name(),
			ChangeOperation.PARAMETER_TYPE_UPDATE.name(),
			ChangeOperation.ATTRIBUTE_TYPE_UPDATE.name()
		);
	
	protected final IsChangeOperationFilter globalRenameFilter = 
		new IsChangeOperationFilter(
			ChangeOperation.ATTRIBUTE_NAME_UPDATE.name(),
			ChangeOperation.METHOD_NAME_UPDATE.name(),
			ChangeOperation.CLASS_NAME_UPDATE.name(),
			ChangeOperation.VARIABLE_NAME_UPDATE.name(), 
			ChangeOperation.PARAMETER_NAME_UPDATE.name()
		);

	protected DiffCatParentMap parentMap;
	
	@Override
	public Set<DiffCatResult> reconcile (Set<DiffCatResult> results)
	{
		Set<DiffCatResult> globalRenames = SetUtil.filterAndCopy(results, globalRenameFilter); 
		Set<DiffCatResult> typeUpdates = SetUtil.filterAndCopy(results, typeFilter);
		
		removeTypeUpdates(typeUpdates, globalRenames);
		
		for (DiffCatResult result : results) {
			if (onlyAffectedByRenames(result, globalRenames)) {
				result.setChangeType(ChangeOperation.AFFECTED_ONLY_BY_RENAMES.name());
			}
		}
		
		return results;
	}
	
	protected void removeTypeUpdates (Collection<DiffCatResult> typeUpdates, Collection<DiffCatResult> renames) {
		Set<DiffCatResult> toRetain = new HashSet<DiffCatResult>();
		for (DiffCatResult rename : renames) {
			String changeType = rename.getChangeType();
			if (changeType.equals(ChangeOperation.ATTRIBUTE_NAME_UPDATE.name()) ||
				changeType.equals(ChangeOperation.PARAMETER_NAME_UPDATE.name()) ||
				changeType.equals(ChangeOperation.VARIABLE_NAME_UPDATE.name())
			)
			{
				IVariableBinding leftVar = extractBindingFromName ((SimpleName) rename.getNodeFrom());
				IVariableBinding rightVar = extractBindingFromName ((SimpleName) rename.getNodeTo());
				
				boolean hadTypeUpdate = false;
				for (DiffCatResult typeUpdate : typeUpdates)
				{
					IVariableBinding leftVar2 = extractBindingFromTypeUpdate(typeUpdate.getNodeFrom(), typeUpdate.getChangeType());
					IVariableBinding rightVar2 = extractBindingFromTypeUpdate(typeUpdate.getNodeTo(), typeUpdate.getChangeType());
					
					if (leftVar == leftVar2 && rightVar == rightVar2) {
						hadTypeUpdate = true;
					}
				}
				
				if (! hadTypeUpdate) {
					toRetain.add(rename);
				}
			}
			else
			{
				toRetain.add(rename);
			}
		}
		typeUpdates.retainAll(toRetain);
	}
	
	protected IVariableBinding extractBindingFromName (Name name) {
		return (IVariableBinding) name.resolveBinding();
	}
	
	protected IVariableBinding extractBindingFromTypeUpdate (ASTNode update, String changeType) {
		if (changeType.equals(ChangeOperation.ATTRIBUTE_TYPE_UPDATE.name())) {
			ASTNode node = ASTUtil.loopUntilNodeType(update, ASTNode.FIELD_DECLARATION);
			FieldDeclaration decl = (FieldDeclaration) node;
			return ((VariableDeclarationFragment) decl.fragments().get(0)).resolveBinding();
		}
		else if (changeType.equals(ChangeOperation.PARAMETER_TYPE_UPDATE.name())){
			ASTNode node = ASTUtil.loopUntilNodeType(update, ASTNode.SINGLE_VARIABLE_DECLARATION);
			SingleVariableDeclaration decl = (SingleVariableDeclaration) node;
			return decl.resolveBinding();
		}
		else {
			ASTNode node = ASTUtil.loopUntilNodeType(update, ASTNode.VARIABLE_DECLARATION_STATEMENT);
			VariableDeclarationStatement decl = (VariableDeclarationStatement) node;
			return ((VariableDeclarationFragment) decl.fragments().get(0)).resolveBinding();
		}
	}
	
	private boolean onlyAffectedByRenames (DiffCatResult r, Collection<DiffCatResult> renames) 
	{
		if (r.getChangeType().equals(ChangeOperation.PARENT_CLASS_UPDATE.name()) ||
			r.getChangeType().equals(ChangeOperation.PARENT_INTERFACE_UPDATE.name()))
		{
			Type left = (Type) r.getNodeFrom();
			
			if (left == null) {
				return false;
			}
			
			Type right = (Type) r.getNodeTo();
			
			if (right == null) {
				return false;
			}
			
			ITypeBinding leftB = left.resolveBinding();
			ITypeBinding rightB = right.resolveBinding();
			
			if (rightB == null) {
				return false;
			}
			
			return onlyAffectedByRename(leftB, rightB, renames);
		}
		
		if (r.getChangeType().equals(ChangeOperation.CLASS_DELETE.name())) {
			
			AbstractTypeDeclaration left = (AbstractTypeDeclaration) r.getNodeFrom();
			
			ITypeBinding leftB = left.resolveBinding();
			
			boolean b = false;
			for (DiffCatResult rename : renames) {
				if (b) continue;
				if (rename.getChangeType().equals(ChangeOperation.CLASS_NAME_UPDATE.name()))
				{
					AbstractTypeDeclaration leftD = (AbstractTypeDeclaration) rename.getLeftParent();
					
					ITypeBinding leftB2 = leftD.resolveBinding();
					
					b = leftB == leftB2;
				}
			}
			
			return b;
			
		}
		
		if (r.getChangeType().equals(ChangeOperation.CLASS_INSERT.name())) {
			
			AbstractTypeDeclaration right = (AbstractTypeDeclaration) r.getNodeTo();
			
			ITypeBinding rightB = right.resolveBinding();
			
			boolean b = false;
			for (DiffCatResult rename : renames) {
				if (b) continue;
				if (rename.getChangeType().equals(ChangeOperation.CLASS_NAME_UPDATE.name()))
				{
					AbstractTypeDeclaration rightD = (AbstractTypeDeclaration) rename.getRightParent();
					
					ITypeBinding rightB2 = rightD.resolveBinding();
					
					b = BindingUtil.areEqual(rightB, rightB2);
				}
			}
			
			return b;
		}
		
		if(! canBeInvestigated(r)) {
			return false;
		}

		List<ASTNode> flatFrom = flattenFromNode(r);
		List<ASTNode> flatTo = flattenToNode(r);
		
		return onlyAffectedByRenames (flatFrom, flatTo, renames); 
	}

	private boolean canBeInvestigated (DiffCatResult r) {

		if (! ChangeOperation.isStatementUpdate(r.getChangeType())) {
			return false;
		}
		if (ChangeOperation.VARIABLE_NAME_UPDATE.name().equals(r.getChangeType())) {
			return false;
		}
		if (ChangeOperation.PARAMETER_NAME_UPDATE.name().equals(r.getChangeType())) {
			return false;
		}
		if (ChangeOperation.ATTRIBUTE_NAME_UPDATE.name().equals(r.getChangeType())) {
			return false;
		}
		if (ChangeOperation.CLASS_NAME_UPDATE.name().equals(r.getChangeType())) {
			return false;
		}
		if (ChangeOperation.METHOD_NAME_UPDATE.name().equals(r.getChangeType())) {
			return false;
		}
		if (ChangeOperation.AFFECTED_BY_VAR_EXTRACTION.equals(r.getChangeType())) {
			return false;
		}
		
		return true;
	}
	
	public boolean onlyAffectedByRenames (List<ASTNode> from, List<ASTNode> to, Collection<DiffCatResult> renames)
	{
		// if sizes not equal, means inserted something
		if (from.size() != to.size()) {
			return false;
		}
		
		// check all have same type and same parents
		for (int i = 0; i < from.size(); i++) {
			ASTNode fromNode = from.get(i);
			ASTNode toNode = to.get(i);
			
			if (fromNode.getNodeType() != toNode.getNodeType()) {
				return false;
			}
			
			if (i < from.size() - 1) {
				ASTNode nextFrom = from.get(i+1);
				ASTNode nextTo   = to.get(i+1);
				
				if (nextFrom.getParent() != fromNode && nextTo.getParent() == toNode) {
					return false;
				}
				
				if (nextFrom.getParent() == fromNode && nextTo.getParent() != toNode) {
					return false;
				}
			}
		}
		
		// now check all 
		for (int i = 0; i < from.size(); i++) {
			ASTNode fromNode = from.get(i);
			ASTNode toNode = to.get(i);
			
			// current node is child
			if (i == from.size() - 1 || (from.get(i+1).getParent() != fromNode))
			{
				if (! fromNode.toString().trim().equals(toNode.toString().trim())) {
					if (fromNode.getNodeType() != ASTNode.SIMPLE_NAME) {
						return false;
					}
					
					SimpleName fromName = (SimpleName) fromNode;
					SimpleName toName = (SimpleName) toNode;
					
					if (! onlyAffectedByRename(fromName, toName, renames)) {
						return false;
					}
				}
			}
		}
		
		return true;
	}
	
	private boolean onlyAffectedByRename (SimpleName left, SimpleName right, Collection<DiffCatResult> renames)
	{
		IBinding leftBinding = left.resolveBinding();
		IBinding rightBinding = right.resolveBinding();
		
		if (leftBinding == null || rightBinding == null) {
			return false;
		}
		
		if (leftBinding.getKind() == IBinding.VARIABLE && rightBinding.getKind() == IBinding.VARIABLE) {
			return onlyAffectedByRename ((IVariableBinding) leftBinding, (IVariableBinding) rightBinding, renames);
		}
		
		if (leftBinding.getKind() == IBinding.METHOD && rightBinding.getKind() == IBinding.METHOD) {
			return onlyAffectedByRename ((IMethodBinding) leftBinding, (IMethodBinding) rightBinding, renames);
		}
		
		if (leftBinding.getKind() == IBinding.TYPE && rightBinding.getKind() == IBinding.TYPE) {
			return onlyAffectedByRename ((ITypeBinding) leftBinding, (ITypeBinding) rightBinding, renames);
		}
		
		return false;
	}
	
	private boolean onlyAffectedByRename (IVariableBinding left, IVariableBinding right, Collection<DiffCatResult> renames)
	{
		if (left.isField() != right.isField() || left.isParameter() != right.isParameter()) {
			return false;
		}
		
		boolean result = false;
		for (DiffCatResult r : renames) {
			
			if (result) {
				continue;
			}
			
			if (r.getChangeType().equals(ChangeOperation.ATTRIBUTE_NAME_UPDATE.name())) {
				
				SimpleName nameFrom = (SimpleName) r.getNodeFrom();
				SimpleName nameTo = (SimpleName) r.getNodeTo();
								
				IVariableBinding leftFragBinding = (IVariableBinding) nameFrom.resolveBinding();
				IVariableBinding rightFragBinding = (IVariableBinding) nameTo.resolveBinding();
				
				result = leftFragBinding == left && rightFragBinding == right;
			} 
			else if (r.getChangeType().equals(ChangeOperation.PARAMETER_NAME_UPDATE.name())) {
				
				Name leftDecl = (Name) r.getNodeFrom();
				Name rightDecl = (Name) r.getNodeTo();
				
				IVariableBinding leftBinding = (IVariableBinding) leftDecl.resolveBinding();
				IVariableBinding rightBinding = (IVariableBinding) rightDecl.resolveBinding();
				
				result = leftBinding == left && rightBinding == right;
			} 
			else if (r.getChangeType().equals(ChangeOperation.VARIABLE_NAME_UPDATE.name())) {
				
				IVariableBinding binding = null;
				
				if (r.getNodeFrom().getNodeType() == ASTNode.VARIABLE_DECLARATION_FRAGMENT) {
					VariableDeclarationFragment leftDecl = (VariableDeclarationFragment) r.getNodeFrom();
					binding = leftDecl.resolveBinding();
				}
				else if (r.getNodeFrom().getNodeType() == ASTNode.VARIABLE_DECLARATION_STATEMENT) {
					VariableDeclarationStatement leftdecl = (VariableDeclarationStatement) r.getNodeFrom();
					VariableDeclarationFragment leftDecl = (VariableDeclarationFragment) leftdecl.fragments().get(0);
					binding = leftDecl.resolveBinding();
				}
				else if (r.getNodeFrom().getNodeType() == ASTNode.SIMPLE_NAME) {
					binding = (IVariableBinding) ((SimpleName) r.getNodeFrom()).resolveBinding();
				}
				else {
					SingleVariableDeclaration leftDecl = (SingleVariableDeclaration) r.getNodeFrom();
					binding = leftDecl.resolveBinding();
				}
				
				IVariableBinding binding2 = null;
				
				if (r.getNodeTo().getNodeType() == ASTNode.VARIABLE_DECLARATION_FRAGMENT) {
					VariableDeclarationFragment rightDecl = (VariableDeclarationFragment) r.getNodeTo();
					binding2 = rightDecl.resolveBinding();
				}
				else if (r.getNodeTo().getNodeType() == ASTNode.VARIABLE_DECLARATION_STATEMENT){
					VariableDeclarationStatement rightdecl = (VariableDeclarationStatement) r.getNodeTo();
					VariableDeclarationFragment rightDecl = (VariableDeclarationFragment) rightdecl.fragments().get(0);
					binding2 = rightDecl.resolveBinding();
				}
				else if (r.getNodeTo().getNodeType() == ASTNode.SIMPLE_NAME) {
					binding2 = (IVariableBinding) ((SimpleName) r.getNodeTo()).resolveBinding();
				}
				else {
					SingleVariableDeclaration rightDecl = (SingleVariableDeclaration) r.getNodeTo();
					binding2 = rightDecl.resolveBinding();
				}
				
				result = binding == left && binding2 == right;
			}
		}
		
		return result;
	}
	
	private boolean onlyAffectedByRename (IMethodBinding left, IMethodBinding right, Collection<DiffCatResult> renames)
	{
		boolean result = false;
		for (DiffCatResult r : renames) {
			if (result) {
				continue;
			}
			
			if (r.getChangeType().equals(ChangeOperation.METHOD_NAME_UPDATE.name())) {
				MethodDeclaration leftDecl = (MethodDeclaration) r.getLeftParent();
				MethodDeclaration rightDecl = (MethodDeclaration) r.getRightParent();
				
				IMethodBinding leftBinding = leftDecl.resolveBinding();
				IMethodBinding rightBinding = rightDecl.resolveBinding();
				
				result = leftBinding == left && rightBinding == right;
			}
			
			// we don't record constructor renames
			if (r.getChangeType().equals(ChangeOperation.CLASS_NAME_UPDATE.name()) && left.isConstructor() && right.isConstructor()) {
				AbstractTypeDeclaration leftDecl = (AbstractTypeDeclaration) r.getLeftParent();
				AbstractTypeDeclaration rightDecl = (AbstractTypeDeclaration) r.getRightParent();
				
				ITypeBinding leftBinding = leftDecl.resolveBinding();
				ITypeBinding rightBinding = rightDecl.resolveBinding();
				
				// because constructors have same name as their declaring class
				ITypeBinding leftName = left.getDeclaringClass();
				ITypeBinding rightName = right.getDeclaringClass();
				
				result = leftBinding == leftName && rightBinding == rightName;
			}
		}
		
		return result;
	}
	
	public boolean onlyAffectedByRename (ITypeBinding left, ITypeBinding right, Collection<DiffCatResult> renames)
	{
		boolean result = false;
		for (DiffCatResult r : renames) {
			if (result) {
				continue;
			}
			
			if (r.getChangeType().equals(ChangeOperation.CLASS_NAME_UPDATE.name())) {
				AbstractTypeDeclaration leftDecl = (AbstractTypeDeclaration) r.getLeftParent();
				AbstractTypeDeclaration rightDecl = (AbstractTypeDeclaration) r.getRightParent();
				
				ITypeBinding leftBinding = leftDecl.resolveBinding();
				ITypeBinding rightBinding = rightDecl.resolveBinding();
				
				result = leftBinding == left && rightBinding == right;
			}
		}
		
		return result;
	}
}
