/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.reconciler;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Statement;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.refiner.AbstractRefiner;
import ca.mcgill.cs.swevo.diffcat.refiner.ThisExpressionRefiner;
import ca.mcgill.cs.swevo.diffcat.util.DiffCatParentMap;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LRTuple;

public class SmallMethodReconciler extends AbstractReconciler {

	private final AbstractRefiner thisExpr = new ThisExpressionRefiner();
	
	@Override
	public Set<DiffCatResult> reconcile(Set<DiffCatResult> results) {
		
		DiffCatParentMap map = new DiffCatParentMap(results);
		for (LRTuple<ASTNode> methodKeys : map.getMethodKeys())
		{
			Collection<DiffCatResult> collection = map.get(methodKeys);
			if (collection.size() != 2) {
				continue;
			}
			
			if (! isSmallMethodPair(methodKeys)) {
				continue;
			}
			
			DiffCatResult del=null, ins=null;
			for (DiffCatResult r : collection) {
				if (r.isStatementDelete()) {
					del = r;
				} 
				if (r.isStatementInsert()) {
					ins = r;
				}
			}
			
			if (del == null || ins == null) {
				continue;
			}
			
			// very small method
			DiffCatResult shell = new DiffCatResult();
			shell.setChangeType(ChangeOperation.IF_STATEMENT_UPDATE.name());
			shell.setNodeFrom(del.getNodeFrom());
			shell.setNodeTo(ins.getNodeTo());
			
			// identical statement is impossible, cuz we woulda seen it
			// so check for this.
			Set<DiffCatResult> set = thisExpr.refine(shell);
			if (set.size() != 1) {
				continue;
			}
			
			DiffCatResult r = CollectionUtil.getFirst(set);
			if (r.getChangeType().equals(ChangeOperation.TRIVIAL_THIS_EXPRESSION_DELETE.name())) {
				r = del.copy();
				r.setNodeTo(ins.getNodeTo());
				r.setChangeType(ChangeOperation.TRIVIAL_THIS_EXPRESSION_DELETE.name());
				results.remove(ins);
				results.remove(del);
				results.add(r);
			}
			else if (r.getChangeType().equals(ChangeOperation.TRIVIAL_THIS_EXPRESSION_INSERT.name())) {
				r = del.copy();
				r.setNodeTo(ins.getNodeTo());
				r.setChangeType(ChangeOperation.TRIVIAL_THIS_EXPRESSION_INSERT.name());
				results.remove(ins);
				results.remove(del);
				results.add(r);
			}
		}
		
		return results;
	}

	private boolean isSmallMethodPair (LRTuple<ASTNode> pair) {
		MethodDeclaration left = (MethodDeclaration) pair.getLeft();
		MethodDeclaration right = (MethodDeclaration) pair.getRight();
		
		if (isSmallMethod(right) && isSmallMethod(left)) {
			return true;
		}
		
		return false;
	}
	
	private boolean isSmallMethod (MethodDeclaration decl) {
		Block body = decl.getBody();
		if (body == null) {
			return false;
		}
		
		if (body.statements().size() != 1) {
			return false;
		}
		
		Statement stmt = (Statement) body.statements().get(0);
		if (stmt.getNodeType() == ASTNode.IF_STATEMENT || 
			stmt.getNodeType() == ASTNode.WHILE_STATEMENT || 
			stmt.getNodeType() == ASTNode.FOR_STATEMENT || 
			stmt.getNodeType() == ASTNode.ENHANCED_FOR_STATEMENT || 
			stmt.getNodeType() == ASTNode.DO_STATEMENT)
		{
			return false;
		}
		
		return true;
	}
}
