/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.refiner;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.BooleanLiteral;
import org.eclipse.jdt.core.dom.CharacterLiteral;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.NumberLiteral;
import org.eclipse.jdt.core.dom.PostfixExpression;
import org.eclipse.jdt.core.dom.PrefixExpression;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.StringLiteral;
import org.hibernate.hql.CollectionSubqueryFactory;

import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTFlattener;
import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LRTuple;

public abstract class AbstractRefiner implements IDiffRefiner {

	protected ASTFlattener flattener = new ASTFlattener();
	
	protected Set<DiffCatResult> singleton (DiffCatResult result) {
		return Collections.singleton(result);
	}

	protected List<ASTNode> flatten(ASTNode one) {
		one.accept(flattener);
		List<ASTNode> oneFlattened = flattener.getFlattenedAndClear();
		return oneFlattened;
	}

	protected boolean areSame(ASTNode subOne, ASTNode subTwo) {
		if (subOne.getNodeType() == ASTNode.SIMPLE_NAME) {
			SimpleName s1 = (SimpleName) subOne;
			SimpleName s2 = (SimpleName) subTwo;
			
			if (! areEqualBindings(s1, s2)) {
				return false;
			}
		}
		else if (subOne.getNodeType() == ASTNode.STRING_LITERAL) {
			StringLiteral lit1 = (StringLiteral) subOne;
			StringLiteral lit2 = (StringLiteral) subTwo;
			
			String s1 = lit1.getLiteralValue();
			String s2 = lit2.getLiteralValue();
			
			if(! s1.equals(s2)) {
				return false;
			}
		}
		else if (subOne.getNodeType() == ASTNode.CHARACTER_LITERAL) {
			CharacterLiteral lit1 = (CharacterLiteral) subOne;
			CharacterLiteral lit2 = (CharacterLiteral) subTwo;
			
			if (! lit1.toString().equals(lit2.toString())) {
				return false;
			}
		}
		else if (subOne.getNodeType() == ASTNode.BOOLEAN_LITERAL) {
			BooleanLiteral lit1 = (BooleanLiteral) subOne;
			BooleanLiteral lit2 = (BooleanLiteral) subTwo;
			
			if (lit1.toString().equals(lit2.toString())) {
				return false;
			}
		}
		else if (subOne.getNodeType() == ASTNode.NUMBER_LITERAL) {
			NumberLiteral lit1 = (NumberLiteral) subOne;
			NumberLiteral lit2 = (NumberLiteral) subTwo;
			
			ITypeBinding t1 = lit1.resolveTypeBinding();
			ITypeBinding t2 = lit2.resolveTypeBinding();
			
			if (! BindingUtil.areEqual(t1, t2)) {
				return false;
			}
			
			if (! lit1.toString().equals(lit2.toString())) {
				return false;
			}
		}
		else if (subOne.getNodeType() == ASTNode.INFIX_EXPRESSION) {
			InfixExpression e1 = (InfixExpression) subOne;
			InfixExpression e2 = (InfixExpression) subTwo;
			
			if (! e1.getOperator().equals(e2.getOperator())) {
				return false;
			}
		}
		else if (subOne.getNodeType() == ASTNode.POSTFIX_EXPRESSION) {
			PostfixExpression p1 = (PostfixExpression) subOne;
			PostfixExpression p2 = (PostfixExpression) subTwo;
			
			if (! p1.getOperator().equals(p2.getOperator())) {
				return false;
			}
		}
		else if (subOne.getNodeType() == ASTNode.PREFIX_EXPRESSION) {
			PrefixExpression p1 = (PrefixExpression) subOne;
			PrefixExpression p2 = (PrefixExpression) subTwo;
			
			if (! p1.getOperator().equals(p2.getOperator())) {
				return false;
			}
		}
		
		return true;
	}

	protected boolean areEqualBindings(SimpleName n1, SimpleName n2) {
		String s1 = n1.toString();
		String s2 = n2.toString();
		
		if (! s1.equals(s2)) {
			return false;
		}
		
		IBinding b1 = n1.resolveBinding();
		IBinding b2 = n2.resolveBinding();
		
		if (b1 == null || b2 == null) {
			return false;
		}
		
		if (b1.getKind() != b2.getKind()) {
			return false;
		}
		
		if (b1.getKind() == IBinding.VARIABLE) {
			IVariableBinding v1 = (IVariableBinding) b1;
			IVariableBinding v2 = (IVariableBinding) b2;
			
			// all we need to know in our context
			return v1.isField() == v2.isField();
			
		}
		else if (b1.getKind() == IBinding.METHOD) {
			
			IMethodBinding m1 = (IMethodBinding) b1;
			IMethodBinding m2 = (IMethodBinding) b2;
			
			if (m1.getParameterTypes().length != m2.getParameterTypes().length) {
				return false;
			} 
			
			return true;
		}
		else {
			return true;
		}
	}
}
