/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.diffcat.util.ast;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.EnumDeclaration;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;

import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.ppa.ValidatorUtil;
import ca.mcgill.cs.swevo.util.ArrayUtil;

public class ASTUtil {

	private static String DOT = ".";
	public static String NON_REGEX_DOT = "\\.";
	
	/**
	 * Converts the fully qualified name into a single string
	 * @param fqn the fully qualified name
	 * @return a dot-separated representation of fqn
	 */
	public static String convertFQN (String[] fqn) {
		StringBuffer buffer = new StringBuffer();
		
		if (fqn.length > 0) {
			buffer.append(fqn[0]);
		}
		
		for (int i = 1; i < fqn.length; i++) {
			buffer.append(DOT);
			buffer.append(fqn[i]);
		}
		
		return buffer.toString();
	}
	
	public static String[] getTypeFQNViaStrings (SingleVariableDeclaration declaration) {
		Type type = declaration.getType();
		String typeString = type.toString();
		typeString = BindingUtil.removeGenerics(typeString);
		return typeString.split(NON_REGEX_DOT);
	}
	
	public static String[] getTypeFQN (SingleVariableDeclaration declaration){ 
		IVariableBinding binding = declaration.resolveBinding();
		if (binding != null) {
			ITypeBinding typeBinding = binding.getType();
			
			if (typeBinding != null) {
				return BindingUtil.getTypeFQN (typeBinding);
			}
		} 
		
		Type type = declaration.getType();
		
		if (type == null) {
			return BindingUtil.UNKNOWN_TYPE_BINDING;
		}
		
		ITypeBinding typeBinding = type.resolveBinding();
		
		if (typeBinding == null) {
			return BindingUtil.UNKNOWN_TYPE_BINDING;
		}
		
		return BindingUtil.getTypeFQN(typeBinding);
	}
	
	
	
	public static ASTNode getSuperType (AbstractTypeDeclaration declaration) {
		
		if (declaration instanceof TypeDeclaration) {
			return getSuperType ((TypeDeclaration) declaration);
		}
		
		return null;
	}
	
	public static ASTNode getSuperType (TypeDeclaration declaration) {
		
		if(declaration.getSuperclassType() != null) {
			return declaration.getSuperclassType();
		}
		
		return null;
	}
	
	public static String getMethodName (MethodDeclaration decl) {
		return decl.getName().getFullyQualifiedName();
	}
	
	@SuppressWarnings({ "rawtypes" })
	public static String[][] getParamTypesViaBindings (MethodDeclaration declaration) {
		List params = declaration.parameters();
		String[][] types = new String[params.size()][];
		
		for (int i = 0; i < types.length; i++) 
		{
			SingleVariableDeclaration var = (SingleVariableDeclaration) params.get(i);
			
			if (var.resolveBinding() != null) {
				ITypeBinding binding = var.resolveBinding().getType();
				if (binding != null) {
					types[i] = BindingUtil.getTypeNoGenericsFQN(binding);
				}
				else {
					binding = var.getType().resolveBinding();
					if (binding != null) {
						types[i] = BindingUtil.getTypeNoGenericsFQN(binding);
					}
				}
			}
			
			if (types[i] == null) {
				types[i] = BindingUtil.UNKNOWN_TYPE_BINDING;
			}
		}
		
		return types;
	}
	
	@SuppressWarnings("rawtypes")
	public static String[][] getParamTypesViaStrings (MethodDeclaration declaration) {
		List params = declaration.parameters();
		String[][] types = new String[params.size()][];
		
		for (int i = 0; i < types.length; i++) 
		{
			SingleVariableDeclaration var = (SingleVariableDeclaration) params.get(i);
			Type type = var.getType();
			String argString = type.toString();
			
			argString = BindingUtil.removeGenerics (argString);
			argString = BindingUtil.translateVarArgs(argString);
			
			types[i] = argString.split(NON_REGEX_DOT);
		}
		
		return types;
	}
	
	@SuppressWarnings("rawtypes")
	public static List<AbstractTypeDeclaration> getTypeDeclarations (CompilationUnit unit) {
		List types = unit.types();
		List<AbstractTypeDeclaration> decls = new LinkedList<AbstractTypeDeclaration>();
		for (Object o : types) {
			decls.add((AbstractTypeDeclaration) o);
		}
		return decls;
	}
	
	@SuppressWarnings("rawtypes")
	public static String[][] getParamTypesIgnoreBrackets (MethodDeclaration declaration) {
		List params = declaration.parameters();
		String[][] types = new String[params.size()][];
		
		for (int i = 0; i < types.length; i++) 
		{
			SingleVariableDeclaration var = (SingleVariableDeclaration) params.get(i);
			
			if (var.resolveBinding() != null) {
				ITypeBinding binding = var.resolveBinding().getType();
				if (binding != null) {
					types[i] = BindingUtil.getTypeFQN(binding);
				}
				else {
					binding = var.getType().resolveBinding();
					if (binding != null) {
						types[i] = BindingUtil.getTypeFQN(binding);
					}
				}
			}
			
			if (types[i] == null) {
				types[i] = new String[]{"UNKNOWN"};
			}
		}
		
		return types;
	}
	
	@SuppressWarnings("rawtypes")
	public static String[] getParamNames (MethodDeclaration declaration) {
		List params = declaration.parameters();
		String[] names = new String[params.size()];
		
		for (int i = 0; i < names.length; i++)
		{
			SingleVariableDeclaration var = (SingleVariableDeclaration) params.get(i);
			names[i] = var.getName().getFullyQualifiedName();
		}
			
		return names;
	}
	
	public static SingleVariableDeclaration getParameter (int i, MethodDeclaration declaration) {
		return (SingleVariableDeclaration) declaration.parameters().get(i);
	}
	
	public static SingleVariableDeclaration getParameter (String name, MethodDeclaration declaration) {
		String[] names = getParamNames(declaration);
		int index = ArrayUtil.indexOf(names, name);
		
		if (index == -1) {
			return null;
		}
		
		return getParameter (index, declaration);
	}
	
	public static String getFieldDeclName (FieldDeclaration declaration) {
		return getDeclName(declaration.fragments());
	}
	
	public static String getVarDeclName (VariableDeclarationStatement declaration) {
		return getDeclName (declaration.fragments());
	}
	
	@SuppressWarnings("rawtypes")
	private static String getDeclName (List fragments) {
		if (fragments.isEmpty()) {
			return "";
		}
		
		VariableDeclarationFragment var = (VariableDeclarationFragment) fragments.get(0);
		
		if (var == null) {
			return "";
		}
		
		return var.getName().getFullyQualifiedName();
	}
	
	@SuppressWarnings("rawtypes")
	public static SimpleName getFieldDeclNameNode (FieldDeclaration declaration) {
		
		List fragments = declaration.fragments();
		
		return ((VariableDeclarationFragment) fragments.get(0)).getName();
	}
	
	@SuppressWarnings("rawtypes")
	public static List getInterfaceTypes (AbstractTypeDeclaration declaration)
	{
		if (declaration.getNodeType()==ASTNode.TYPE_DECLARATION) {
			TypeDeclaration typeDeclaration = (TypeDeclaration) declaration;
			return typeDeclaration.superInterfaceTypes();
		}
		else if (declaration.getNodeType()==ASTNode.ENUM_DECLARATION) {
			EnumDeclaration enumDeclaration = (EnumDeclaration) declaration;
			return enumDeclaration.superInterfaceTypes();
		}
		
		return Collections.EMPTY_LIST;
	}
	
	public final static String TYPE_SEPARATOR = "$";

	/**
	 * Returns the AbstractTypeDeclaration which declares the given container.
	 */
	public static ASTNode getTypeContainer(ASTNode node) {
		ASTNode container = null;

		if (node == null) {
			container = null;
		} else if (isTypeContainer(node)) {
			container = (AbstractTypeDeclaration) node;
		} else {
			container = getTypeContainer(node.getParent());
		}

		return container;
	}
	
	public static ASTNode loopUntilNodeType (ASTNode node, int kind) {
		if (node == null) {
			return null;
		}
		if (node.getNodeType() == kind) {
			return node;
		}
		
		return loopUntilNodeType (node.getParent(), kind); 
	}
	
	
	
	/**
	 * Returns the next parent of the current node with the appropriate type
	 */
	public static ASTNode getParentWithType (int type, ASTNode node) {
		if (node == null) {
			return null;
		}
		ASTNode parent = node.getParent();
		while (parent != null && parent.getNodeType() != type) {
			parent = parent.getParent();
		}
		
		return parent;
	}
	
	public static ASTNode getCompilationUnit (ASTNode node) {
		if (node == null) {
			return null;
		}
		else if (node instanceof CompilationUnit) {
			return node;
		} 
		else {
			return getCompilationUnit(node.getParent());
		}
	}
	
	public static boolean isTypeContainer(ASTNode node) {
		return node != null && (node instanceof AbstractTypeDeclaration || node instanceof AnonymousClassDeclaration);
	}
	
	public static String getSimpleNameFromFQN(String name) {
		ValidatorUtil.validateEmpty(name, "name", true);
		String simpleName = name;
		int index = name.lastIndexOf(".");
		if (index != -1 && index != (name.length() - 1)) {
			simpleName = name.substring(index + 1);
		}

		return simpleName;
	}
	
	public static boolean isLastReturnStatement (MethodDeclaration method, ReturnStatement statement) {
		List<ReturnStatement> stmts = getReturnStatements(method);
		if (stmts.isEmpty()) {
			return false;	// strange
		}
		return stmts.get(stmts.size() - 1) == statement;
	}
	
	public static List<ReturnStatement> getReturnStatements (MethodDeclaration method) {
		ReturnStatementCollector col = new ReturnStatementCollector();
		method.accept(col);
		return col.getStmts();
	}
	
	private static class ReturnStatementCollector extends ASTVisitor {
		private List<ReturnStatement> stmts = new ArrayList<ReturnStatement>();
		public boolean visit (ReturnStatement stmt) {
			stmts.add(stmt);
			return false;
		}
		public List<ReturnStatement> getStmts () {
			return Collections.unmodifiableList(stmts);
		}
	}
	
	public static IBinding getBinding (ASTNode node) {
		if (node instanceof AbstractTypeDeclaration) {
			return ((AbstractTypeDeclaration) node).resolveBinding();
		}
		if (node instanceof MethodDeclaration) {
			return ((MethodDeclaration) node).resolveBinding();
		}
		if (node instanceof FieldDeclaration) {
			return ((VariableDeclarationFragment) ((FieldDeclaration) node).fragments().get(0)).resolveBinding();
		}
		if (node instanceof VariableDeclarationFragment) {
			return ((VariableDeclarationFragment) node).resolveBinding();
		}
		
		return null;
	}
	
	@SuppressWarnings("rawtypes")
	public static Set<ASTNode> getDeclaredTypes (CompilationUnit unit) {
		
		List types = unit.types();
		
		Set<ASTNode> results = new HashSet<ASTNode>();
		
		for (int i = 0; i < types.size(); i++) {
			AbstractTypeDeclaration abstractTypeDeclaration = (AbstractTypeDeclaration) types.get(i);
			results.add(abstractTypeDeclaration);
			results.addAll(ASTUtil.getMemberTypes(abstractTypeDeclaration));	// to pick up named inner classes
		}
		
		return results;
	}
	
	public static Set<ASTNode> getMemberTypes (AbstractTypeDeclaration declaration) {
		
		Set<ASTNode> types = new HashSet<ASTNode>();
		for (Object o : declaration.bodyDeclarations()) {
			if (o instanceof AbstractTypeDeclaration) {
				AbstractTypeDeclaration t = (AbstractTypeDeclaration) o;
				ITypeBinding binding = t.resolveBinding();
				if (binding.isMember()) {
					types.add(t);
				}
			}
		}
		
		return types;
	}
	
	@SuppressWarnings("rawtypes")
	public static List<MethodDeclaration> getConstructors (AbstractTypeDeclaration declaration) {
		
		List bodyDeclarations = declaration.bodyDeclarations();
		List<MethodDeclaration> result = new ArrayList<MethodDeclaration>();
		
		for (Object o : bodyDeclarations) {
			if (o instanceof MethodDeclaration) {
				MethodDeclaration o2 = (MethodDeclaration) o;
				if (o2.resolveBinding().isConstructor()) {
					result.add(o2);
				}
			}
		}
		
		return result;
	}
	
	public static List<BodyDeclaration> getFieldsAndMethods (AbstractTypeDeclaration declaration) {
		
		List<BodyDeclaration> result = new ArrayList<BodyDeclaration>();
		result.addAll(getMethods(declaration));
		result.addAll(getFields(declaration));
		return result;
	}
	
	@SuppressWarnings("rawtypes")
	public static List<MethodDeclaration> getMethods (AbstractTypeDeclaration declaration) {
		
		List bodyDeclarations = declaration.bodyDeclarations();
		List<MethodDeclaration> methods = new ArrayList<MethodDeclaration> ();
		
		for (Object o : bodyDeclarations) {
			if (o instanceof MethodDeclaration) {
				methods.add((MethodDeclaration) o);
			}
		}
		
		return methods;
	}
	
	@SuppressWarnings("rawtypes")
	public static List<FieldDeclaration> getFields (AbstractTypeDeclaration declaration) {
		
		List bodyDeclarations = declaration.bodyDeclarations();
		List<FieldDeclaration> fields = new ArrayList<FieldDeclaration> ();
		
		for (Object o : bodyDeclarations) {
			if (o instanceof FieldDeclaration) {
				fields.add((FieldDeclaration) o);
			}
		}
		
		return fields;
	}
}
