/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ca.mcgill.cs.swevo.util.filter.IEqualityFilter;
import ca.mcgill.cs.swevo.util.filter.IFilter;
import ca.mcgill.cs.swevo.util.op.DownCaster;
import ca.mcgill.cs.swevo.util.op.IOperation;

/**
 * Helper utilities to deal with collections.
 * @author david
 *
 */
public class CollectionUtil {
	
	/**
	 * Returns the first element found in the collection
	 */
	public static <T> T getFirst (Collection<T> collection) {
		
		T first = null;
		for (T t : collection) {
			if (first == null) {
				first = t;
			}
		}
		
		return first;
	}
	
	public static <T> T getNth (Collection<T> collection, int n) {
		T nth = null;
		int i = 1;
		for (T t : collection) {
			if (i++ == n) {
				nth = t;
			}
		}
		return nth;
	}
	
	/**
	 * Retains only those elements accepted by the filter and returns
	 * a new collection containing those elements
	 */
	public static <T> Collection<T> filterAndCopy (Collection<T> collection, IFilter<T> filter) {
		Collection<T> retain = new ArrayList<T>(collection.size());
		for (T item : collection) {
			boolean passes = filter.passesFilter(item); 
			if (passes) {
				retain.add(item);
			}
		}
		return retain;
	}
	
	/**
	 * Returns true if the given element passes all the filters
	 */
	public static <T> boolean passesFilter (T element, Collection<IFilter<T>> filters) {
		boolean yes = true;
		for (IFilter<T> f : filters) {
			yes &= f.passesFilter(element);
		}
		return yes;
	}
	
	/**
	 * Retains only those elements that pass all filters;
	 */
	public static <T> Collection<T> filterAndCopy (Collection<T> collection, Collection<IFilter<T>> filters) {
		Collection<T> retain = new ArrayList<T>(collection.size());
		for (T item : collection) {
			boolean passes = true;
			for (IFilter<T> filter : filters)
				passes &= passes? filter.passesFilter(item) : false;
			if (passes) retain.add(item);
		}
		return retain;
	}
	
	/**
	 * Returns true if the collection contains at least one element matching the filter
	 */
	public static <T> boolean contains (Collection<T> collection, IFilter<? super T> filter) {
		boolean contains = false;
		for (T item : collection) {
			if (! contains) {
				contains = filter.passesFilter(item);
			}
		}
		return contains;
	}
	
	public static <T> boolean contains (Collection<T> collection, IEqualityFilter<? super T> filter) {
		boolean contains = false;
		for (T item : collection) {
			if (! contains) {
				contains = filter.passesFilter(item);
			}
		}
		return contains;
	}
	
	public static <T> boolean contains (Collection<T> collection, Class<? extends T> c) {
		boolean contains = false;
		for (T item : collection) {
			if (! contains) contains = (item.getClass() == c);
		}
		return contains;
	}
	
	/**
	 * Applies the given operation to all elements in the collection. 
	 * This is different from conversion
	 */
	public static <T> void applyToAll (Collection<T> collection, IOperation<T> op) {
		for (T item : collection) {
			op.apply(item);
		}
	}
	
	@SuppressWarnings("unchecked")
	public static <T> LRTupleCollection<T> getUnorderedPairs (Collection<T> col) {
		Object[] array = col.toArray();
		LRTupleCollection<T> pairs = new LRTupleCollection<T>();
		for (int i = 0; i < array.length; i++) {
			for (int j = i+1; j < array.length; j++) {
				pairs.add((T) array[i], (T) array[j]);
			}
		}
		
		return pairs;
	}
	
	@SuppressWarnings("unchecked")
	public static <T> LRTupleCollection<T> getOrderedPairs (Collection<T> col) {
		Object[] array = col.toArray();
		LRTupleCollection<T> pairs = new LRTupleCollection<T>();
		for (int i = 0; i < array.length; i++) {
			for (int j = 0; j < array.length; j++) {
				if (i==j) continue;
				pairs.add((T) array[i], (T) array[j]);
			}
		}
		return pairs;
	}
	
	public static <T> Collection<T> removeAll (Collection<T> col, IFilter<T> removeCondition) {
		Collection<T> newCol = new ArrayList<T>();
		for (T el : col) {
			if (! removeCondition.passesFilter(el)) {
				newCol.add(el);
			}
		}
		col.retainAll(newCol);
		return col;
	}
	
	public static <T> Collection<T> removeAll (Collection<T> col, Class<? extends T> c) {
		Collection<T> newCol = new ArrayList<T>();
		for (T el : col) {
			if (el.getClass() == c) {
				newCol.add(el);
			}
		}
		return newCol;
	}
	
	public static <F, T extends F> Collection<T> downCast (Collection<F> col, DownCaster<F, T> caster) {
		Collection<T> newCol = new ArrayList<T>();
		for (F el : col) {
			newCol.add(caster.convert(el));
		}
		return newCol;
	}
	
	public static <T> LRTupleSet<T> crossProduct (Collection<T> col1, Collection<T> col2) {
		LRTupleSet<T> crossProduct = new LRTupleSet<T>();
		for (T t1 : col1) {
			for (T t2 : col2) {
				crossProduct.add(t1, t2);
			}
		}
		return crossProduct;
	}
}
 
