/*******************************************************************************
 * Copyright (c) 2011 David Kawrykow.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     David Kawrykow - initial API and implementation
 ******************************************************************************/
package ca.mcgill.cs.swevo.util;

import java.util.Arrays;
import java.util.Collection;
import java.util.Set;

import org.junit.After;
import org.junit.Test;

import static junit.framework.Assert.*;

import ca.mcgill.cs.swevo.util.filter.IEqualityFilter;
import ca.mcgill.cs.swevo.util.filter.IFilter;
import ca.mcgill.cs.swevo.util.op.IConverter;

public class TestCollectionMap {

	private final CollectionMap<String, String> testMap = new CollectionMap<String, String>();
	
	@After
	public void clearMap () {
		testMap.clear();
	}
	
	@Test
	public void testCollectionsModified ()
	{
		buildMap01();
		
		testMap.get("Key1").add("Bob");
		assertTrue(testMap.get("Key1").contains("Bob"));
		
		testMap.get("Key1").remove("Val2");
		assertFalse(testMap.get("Key1").contains("Val2"));
	}
	
	@Test
	public void testRemoveFromKeyWithFilter () 
	{
		buildMap01();
		
		testMap.removeFromKey("Key1", new StringEqFilter("Val1"));
		
		Set<String> keySet = testMap.keySet();
		assertEquals(2, keySet.size());
		assertTrue(keySet.contains("Key1") && keySet.contains("Key2"));
		
		Collection<String> vals = testMap.get("Key1");
		assertEquals(1, vals.size());
		assertTrue(vals.contains("Val2"));
		
		vals = testMap.get("Key2");
		assertEquals(2, vals.size());
		assertTrue(vals.contains("Val1") && vals.contains("Val2"));
	}

	@Test
	public void testTranspose () 
	{
		buildMap01();
		
		CollectionMap<String, String> transpose = testMap.transpose();
		
		Set<String> keySet = transpose.keySet();
		assertEquals(2, keySet.size());
		assertTrue(keySet.contains("Val1") && keySet.contains("Val2"));
		
		Collection<String> vals = transpose.get("Val1");
		assertEquals(3, vals.size());
		assertTrue(vals.contains("Key1") && vals.contains("Key2"));
		
		vals = transpose.get("Val2");
		assertEquals(2, vals.size());
		assertTrue(vals.contains("Key1") && vals.contains("Key2"));
	}
	
	@Test
	public void testConverter ()
	{
		buildMap01();
		
		testMap.applyToKeys(new StringPrefixer());
		
		Set<String> keySet = testMap.keySet();
		assertEquals(2, keySet.size());
		assertTrue(keySet.contains(StringPrefixer.TOKEN + "Key1"));
		assertTrue(keySet.contains(StringPrefixer.TOKEN + "Key2"));
		
		Collection<String> vals = testMap.get(StringPrefixer.TOKEN + "Key1");
		assertEquals(3, vals.size());
		assertTrue(vals.contains("Val1"));
		assertTrue(vals.contains("Val2"));
		assertTrue(vals.contains("Val1"));
		
		
		vals = testMap.get(StringPrefixer.TOKEN + "Key2");
		assertEquals(2, vals.size());
		assertTrue(vals.contains("Val1"));
		assertTrue(vals.contains("Val2"));
	}
	
	@Test
	public void testFilterAll () 
	{
		buildMap02();
		
		testMap.retainAllValues(new StringEqFilter("Val2"));
		
		Collection<String> vals = testMap.get("Key1");
		assertEquals(1, vals.size());
		assertTrue(vals.contains("Val2"));
		
		vals = testMap.get("Key2");
		assertEquals(1, vals.size());
		assertTrue(vals.contains("Val2"));
		
		vals = testMap.get("Key3");
		assertEquals(1, vals.size());
		assertTrue(vals.contains("Val2"));
		
		vals = testMap.get("Key4");
		assertNotNull(vals);
		assertEquals(0, vals.size());
	}
	
	@Test
	public void testRemoveAll () 
	{
		buildMap02();
		
		StringEqFilter values = new StringEqFilter("Val1");
		testMap.removeFromKey("Key1", values);
		testMap.removeFromKey("Key2", values);
		testMap.removeFromKey("Key3", values);
		testMap.removeFromKey("Key4", values);
		
		Collection<String> vals = testMap.get("Key1");
		assertEquals(1, vals.size());
		assertTrue(vals.contains("Val2"));
		
		vals = testMap.get("Key2");
		assertEquals(1, vals.size());
		assertTrue(vals.contains("Val2"));
		
		vals = testMap.get("Key3");
		assertEquals(1, vals.size());
		assertTrue(vals.contains("Val2"));
		
		vals = testMap.get("Key4");
		assertNotNull(vals);
		assertEquals(0, vals.size());
	}
	
	@Test
	public void testRemoveKeyWithFilter ()
	{
		buildMap02();
		
		testMap.removeKey(new StringEqFilter("Key1"));
		
		Set<String> keySet = testMap.keySet();
		assertEquals(3, keySet.size());
		assertFalse(keySet.contains("Key1"));
	}
	
	@Test
	public void testGetValueSet () 
	{
		buildMap01();
		
		Set<String> values = testMap.values();
		assertEquals(2, values.size());
		assertTrue(values.contains("Val1") && values.contains("Val2"));
	}
	
	@Test
	public void testSubMap ()
	{
		buildMap02();
		
		CollectionMap<String,String> subMap = testMap.subMap(Arrays.asList("Key1", "Key2"));
		
		Set<String> keySet = subMap.keySet();
		assertEquals(2, keySet.size());
		assertTrue(keySet.contains("Key1") && keySet.contains("Key2"));
		
		Collection<String> vals = subMap.get("Key1");
		assertEquals(3, vals.size());
		assertTrue(vals.contains("Val1") && vals.contains("Val2"));
		
		testMap.removeFromKey("Key1", new StringEqFilter("Val1"));
		
		vals = subMap.get("Key1");
		assertEquals(3, vals.size());
		assertTrue(vals.contains("Val1") && vals.contains("Val2"));
	}
	
	@Test
	public void testKeyPairSet ()
	{
		buildMap02();
		
		LRTupleSet<String> set = testMap.keyPairSet();
		
		assertEquals(6, set.getSize());
		
	}
	
	@Test
	public void testAddWithFilter ()
	{
		buildMap02();
		
		testMap.add(new StringEqFilter("Key5"), "Hello");
		
		Set<String> keys = testMap.keySet();
		assertTrue(keys.contains("Key5"));
		
		Collection<String> vals = testMap.get("Key5");
		assertEquals(1, vals.size());
		assertTrue(vals.contains("Hello"));
		
		testMap.add(new StringEqFilter("Key1"), "Hello");
		
		vals = testMap.get("Key1");
		assertTrue(vals.contains("Hello"));
		
		vals = testMap.get("Key2");
		assertFalse(vals.contains("Hello"));
	}
	
	
	
	private void buildMap01() {
		testMap.add("Key1", "Val1");
		testMap.add("Key1", "Val2");
		testMap.add("Key1", "Val1");
		testMap.add("Key2", "Val1");
		testMap.add("Key2", "Val2");
	}
	
	private void buildMap02() {
		testMap.add("Key1", "Val1");
		testMap.add("Key1", "Val2");
		testMap.add("Key1", "Val1");
		testMap.add("Key2", "Val1");
		testMap.add("Key2", "Val2");
		testMap.add("Key3", "Val2");
		testMap.add("Key4", "Val1");
	}
	
	public class StringEqFilter implements IEqualityFilter<String> 
	{
		private final String compareTo;
		
		public StringEqFilter (String compareTo) {
			this.compareTo = compareTo;
		}

		@Override
		public boolean passesFilter(String t) {
			return compareTo.equals(t);
		}

		@Override
		public String getComparison() {
			return compareTo;
		}
	}
	
	private static class StringPrefixer implements IConverter<String, String>
	{
		public static String TOKEN = "@";
		
		@Override
		public String convert(String f) {
			return TOKEN + f;
		}
		
	}
}
